import os
import unittest

import rdflib

import sbol2

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
TEST_LOCATION = os.path.join(MODULE_LOCATION, 'resources', 'crispr_example.xml')
PARTS_LOCATION = os.path.join(MODULE_LOCATION, 'resources', 'tutorial', 'parts.xml')


class TestReferencedObjects(unittest.TestCase):

    def test_participant_type(self):
        doc = sbol2.Document()
        doc.read(TEST_LOCATION)
        md_uri = 'http://sbols.org/CRISPR_Example/CRISPR_Template/1.0.0'
        md = doc.moduleDefinitions[md_uri]
        # Work with the first interaction
        i = md.interactions[0]
        # participant should be a str
        self.assertEqual(type(i.participations[0].participant), str)

    def test_fc_definition(self):
        doc = sbol2.Document()
        doc.read(TEST_LOCATION)
        md_uri = 'http://sbols.org/CRISPR_Example/CRISPR_Template/1.0.0'
        md = doc.moduleDefinitions[md_uri]
        fc_uri = ('http://sbols.org/CRISPR_Example/CRISPR_Template' +
                  '/cas9_gRNA_complex/1.0.0')
        fc = md.functionalComponents[fc_uri]
        # definition should be a str
        self.assertEqual(type(fc.definition), str)

    def test_cd_sequences(self):
        # Test a referenced object storing a list instead of a singleton
        doc = sbol2.Document()
        doc.read(PARTS_LOCATION)

        cd_uri = 'http://examples.org/ComponentDefinition/AmeR/1'
        cd = doc.componentDefinitions[cd_uri]

        s1_uri = 'http://examples.org/Sequence/AmeR_sequence/1'
        s2_uri = 'http://examples.org/Sequence/ECK120010818_sequence/1'

        # Ensure the URI is present, and as a string
        self.assertTrue(s1_uri in cd.sequences)

        # Cannot append sequences - it has no effect on the cd
        #
        # The CD returns a copy of the list of sequences, not its
        # internal representation.
        cd.sequences.append(s2_uri)
        self.assertTrue(len(cd.sequences) == 1)

        cd.sequences = [s1_uri, s2_uri]
        self.assertTrue(len(cd.sequences) == 2)

        # Verify that all of the elements are instances of str
        self.assertTrue(all([isinstance(uri, str) for uri in cd.sequences]))

        # Verify that the attribute is still a ReferencedObject and
        # was not overwritten with the list.
        if 'sequences' in cd.__dict__:
            self.assertIsInstance(cd.__dict__['sequences'], sbol2.ReferencedObject)

    def test_dunder_uri(self):
        # See issue #319
        # The __uri__ method to convert an object to a URI
        # was returning a str, not a URIRef
        sa = sbol2.SequenceAnnotation('test_sa')
        comp = sbol2.Component('test_comp')
        sa.component = comp
        expected = rdflib.URIRef(comp.identity)
        self.assertEqual(expected, sa.properties[sbol2.SBOL_COMPONENT_PROPERTY][0])


if __name__ == '__main__':
    unittest.main()



import os
import posixpath
from typing import Union

from rdflib import URIRef

from .component import Component
from .config import Config, ConfigOptions
from .constants import *
from .toplevel import TopLevel
from .property import OwnedObject, ReferencedObject, URIProperty
from .sbolerror import SBOLError, SBOLErrorCode
from .sequence import Sequence
from .sequenceannotation import SequenceAnnotation
from .sequenceconstraint import SequenceConstraint


class ComponentDefinition(TopLevel):
    """
    The ComponentDefinition class represents the structural entities
    of a biological design.

    The primary usage of this class is to represent structural entities
    with designed sequences, such as DNA, RNA, and proteins,
    but it can also be used to represent
    any other entity that is part of a design,
    such as small molecules, proteins, and complexes.
    """

    # The types property is a REQUIRED set of URIs that specifies
    # the category of biochemical or physical entity
    # (for example DNA, protein, or small molecule)
    # that a ComponentDefinition object abstracts
    # for the purpose of engineering design.
    # The types property of every ComponentDefinition
    # MUST contain one or more URIs that MUST identify terms
    # from appropriate ontologies, such as the BioPAX ontology
    # or the ontology of Chemical Entities of Biological Interest.
    # See the table below for examples.
    # | Type              | URI for BioPAX Term                                           | LibSBOL symbol        |  # noqa
    # | :---------------- | :------------------------------------------------------------ | :-------------------- |  # noqa
    # | DNA               | http://www.biopax.org/release/biopax-level3.owl#DnaRegion     | BIOPAX_DNA            |  # noqa
    # | RNA               | http://www.biopax.org/release/biopax-level3.owl#RnaRegion     | BIOPAX_RNA            |  # noqa
    # | Protein           | http://www.biopax.org/release/biopax-level3.owl#Protein       | BIOPAX_PROTEIN        |  # noqa
    # | Small Molecule    | http://www.biopax.org/release/biopax-level3.owl#SmallMolecule | BIOPAX_SMALL_MOLECULE |  # noqa
    # | Complex           | http://www.biopax.org/release/biopax-level3.owl#Complex       | BIOPAX_COMPLEX        |  # noqa
    _types = None  # URIProperty

    # The roles property is an OPTIONAL set of URIs
    # that clarifies the potential function of the entity represented
    # by a ComponentDefinition in a biochemical or physical context.
    # The roles property of a ComponentDefinition MAY
    # contain one or more URIs that MUST identify terms
    # from ontologies that are consistent with the types property
    # of the ComponentDefinition.
    # For example, the roles property of a DNA or RNA ComponentDefinition
    # could contain URIs identifying terms from the Sequence Ontology (SO).
    # See the table below for common examples.
    # | Role              | URI for Sequence Ontology Term            | LibSBOL symbol    |  # noqa
    # | :---------------- | :---------------------------------------- | :---------------- |  # noqa
    # | Miscellaneous     | http://identifiers.org/so/SO:0000001      | SO_MISC           |  # noqa
    # | Promoter          | http://identifiers.org/so/SO:0000167      | SO_PROMOTER       |  # noqa
    # | RBS               | http://identifiers.org/so/SO:0000139      | SO_RBS            |  # noqa
    # | CDS               | http://identifiers.org/so/SO:0000316      | SO_CDS            |  # noqa
    # | Terminator        | http://identifiers.org/so/SO:0000141      | SO_TERMINATOR     |  # noqa
    # | Gene              | http://identifiers.org/so/SO:0000704      |                   |  # noqa
    # | Operator          | http://identifiers.org/so/SO:0000057      |                   |  # noqa
    # | Engineered Gene   | http://identifiers.org/so/SO:0000280      |                   |  # noqa
    # | mRNA              | http://identifiers.org/so/SO:0000234      |                   |  # noqa
    # | Effector          | http://identifiers.org/chebi/CHEBI:35224  |                   |  # noqa
    _roles = None  # URIProperty

    # The components property is OPTIONAL and MAY
    # specify a set of Component objects that are contained
    # by the ComponentDefinition. The components properties
    # of ComponentDefinition objects can be used
    # to construct a hierarchy of Component and ComponentDefinition objects.
    # If a ComponentDefinition in such a hierarchy refers to
    # one or more Sequence objects, and there exist
    # ComponentDefinition objects lower in the hierarchy
    # that refer to Sequence objects with the same encoding,
    # then the elements properties of these Sequence objects
    # SHOULD be consistent with each other, such that well-defined mappings
    # exist from the "lower level" elements
    # to the "higher level" elements. This mapping is also subject
    # to any restrictions on the positions of the Component objects
    # in the hierarchy that are imposed by the SequenceAnnotation
    # or SequenceConstraint objects contained
    # by the ComponentDefinition objects in the hierarchy.
    # The set of relations between Component and ComponentDefinition objects
    # is strictly acyclic.
    components = None  # OwnedObject<Component>

    sequences = None  # ReferencedObject

    sequenceAnnotations = None  # OwnedObject<SequenceAnnotation>

    sequenceConstraints = None  # OwnedObject<SequenceConstraint>

    def __init__(self, uri=URIRef("example"),
                 component_type=URIRef(BIOPAX_DNA),
                 version=VERSION_STRING,
                 type_uri=SBOL_COMPONENT_DEFINITION):
        """Construct a ComponentDefinition.

        :param uri: A full URI including a scheme, namespace, and identifier.
        If SBOLCompliance configuration is enabled, then this argument
        is simply the displayId for the new object
        and a full URI will automatically be constructed.
        :param component_type: A BioPAX ontology term
        that indicates whether the ComponentDefinition
        is DNA, RNA, protein, or some other molecule type.
        :param version: An arbitrary version string.
        If SBOLCompliance is enabled, this should be a Maven version string
        of the form "major.minor.patch".
        :param type_uri: The RDF type for an extension class
        derived from this one
        """
        super().__init__(type_uri, uri, version)
        self.types = URIProperty(self, SBOL_TYPES,
                                 '1', '*', None, component_type)
        self.roles = URIProperty(self, SBOL_ROLES,
                                 '0', '*', None)
        self.sequences = ReferencedObject(self, SBOL_SEQUENCE_PROPERTY,
                                          SBOL_SEQUENCE, '0', '*', None)
        self.sequenceAnnotations = OwnedObject(self, SBOL_SEQUENCE_ANNOTATIONS,
                                               SequenceAnnotation,
                                               '0', '*', None)
        self.components = OwnedObject(self, SBOL_COMPONENTS, Component,
                                      '0', '*', None)
        self.sequenceConstraints = OwnedObject(self, SBOL_SEQUENCE_CONSTRAINTS,
                                               SequenceConstraint,
                                               '0', '*', None)
        # This caches a sequence object so that it can be added
        # to a document later
        self._sequence_cache: Union[Sequence, None] = None

    @property
    def sequence(self):
        seqs = self.sequences
        if not seqs:
            return None
        if self.doc:
            # In a document we always look up the sequence in case
            # the sequence has been removed.
            seq_uri = seqs[0]
            try:
                return self.doc.sequences[seq_uri]
            except SBOLError as e:
                if e.error_code() != SBOLErrorCode.NOT_FOUND_ERROR:
                    raise
                return None
        else:
            # Not in a document, try to use the cache if available
            if self._sequence_cache and self._sequence_cache.identity in seqs:
                return self._sequence_cache
            return None

    @sequence.setter
    def sequence(self, sequence: Union[Sequence, None]):
        if not sequence:
            # Unsetting the sequence
            self.sequences = None
            self._sequence_cache = None
            return
        if self.doc:
            try:
                self.doc.add(sequence)
            except SBOLError as e:
                if e.error_code() != SBOLErrorCode.DUPLICATE_URI_ERROR:
                    raise
        else:
            # Not in a document. Cache the sequence in case we get added
            # to a document later.
            self._sequence_cache = sequence
        self.sequences = [sequence.identity]

    def _added_to_document(self, doc):
        super()._added_to_document(doc)
        # Add the sequence to the document
        if self._sequence_cache:
            try:
                doc.add(self._sequence_cache)
            except SBOLError as e:
                if e.error_code() == SBOLErrorCode.SBOL_ERROR_URI_NOT_UNIQUE:
                    # The sequence is already in the document, this is ok.
                    pass
                else:
                    raise

    def addType(self, new_type):
        val = self.types
        val.append(new_type)
        self.types = val

    def removeType(self, index=0):
        val = self.types
        del val[index]
        self.types = val

    def addRole(self, new_role):
        val = self.roles
        val.append(new_role)
        self.roles = val

    def removeRole(self, index=0):
        val = self.roles
        del val[index]
        self.roles = val

    def assemble(self, component_list, assembly_method=None, doc=None):
        """Assembles ComponentDefinitions into an abstraction hierarchy.

        The resulting data structure is a partial design,
        still lacking a primary structure or explicit sequence.
        To form a primary structure out of the ComponentDefinitions,
        call linearize after calling assemble.
        To fully realize the target sequence, use Sequence::assemble().
        :param component_list: Either a list of URIs
        for the constituent ComponentDefinitions or a list of subcomponents.
        A list of displayIds is also acceptable if using SBOL-compliant URIs.
        :param assembly_method: An optional argument
        such as IGEM_STANDARD_ASSEMBLY that affects how components
        are composed and the final target sequence.
        :param doc: The Document to which the assembled ComponentDefinitions
        will be added. If not set, then you must add this ComponentDefinition
        to a Document before calling this method.
        :return: None
        """

        # Due to the recursive nature of this routine, it is hard to completely
        # validate that all the necessary preconditions for successful execution are met
        # prior to execution. That means if a call fails, it may result in a modified
        # and incomplete data structure that will be difficult to fix when the user is
        # working interactively in the interpreter
        if not Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS):
            raise EnvironmentError('Assemble method requires SBOL-compliance enabled')

        # Validate doc
        if not self.doc and not doc:
            raise ValueError('Missing doc argument. If the ComponentDefinition does '
                             'not belong to a Document, a target Document must be '
                             'specified using the doc keyword argument.')
        if doc and self.doc != doc:
            raise ValueError('Invalid doc argument. Do not use the doc keyword '
                             'argument if the ComponentDefinition already belongs '
                             'to a Document')

        # Validate component_list
        doc = doc if doc else self.doc
        if isinstance(component_list, list) and all(isinstance(c, ComponentDefinition)
                                                    for c in component_list):
            for cdef in component_list:
                if cdef.doc and cdef.doc is not doc:
                    raise ValueError('Invalid component_list specified. Assembly '
                                     'subcomponents must belong to the same Document '
                                     'as self.')
        elif isinstance(component_list, list) and all(isinstance(c, str)
                                                      for c in component_list):
            component_identities = component_list[:]
            component_list = []
            for c_id in component_identities:
                if c_id not in doc.componentDefinitions:
                    raise ValueError('Invalid component_list specified. '
                                     'ComponentDefinition <%s> not found.' % c_id)
                cdef = doc.componentDefinitions[c_id]
                component_list.append(cdef)
        else:
            raise TypeError('Invalid component_list specified. Please provide a list '
                            'of ComponentDefinitions or, alternatively, a list of '
                            'ComponentDefinition displayIds')

        if not self.doc:
            doc.addComponentDefinition(self)
        for cdef in component_list:
            if not cdef.doc:
                self.doc.addComponentDefinition(cdef)

        if assembly_method:
            component_list = assembly_method(component_list)
            if not all(type(c) is ComponentDefinition for c in component_list):
                raise TypeError('Invalid callback specified for assembly_method. The '
                                'callback must return a list of ComponentDefinitions')

        # Instantiate a Component for each ComponentDefinition in the list
        instance_list = []
        for cdef in component_list:

            # Generate URI of new Component.  Check if an object with that URI is
            # already instantiated.
            instance_count = 0
            component_id = self.persistentIdentity + "/" + cdef.displayId + "_" \
                + str(instance_count) + "/" + self.version
            while self.find(component_id):
                # Find the last instance assigned
                instance_count += 1
                component_id = self.persistentIdentity + "/" + cdef.displayId + "_" \
                    + str(instance_count) + "/" + self.version

            c = self.components.create(cdef.displayId + "_" + str(instance_count))
            c.definition = cdef.identity
            instance_list.append(c)
        return component_list

    def assemblePrimaryStructure(self, primary_structure, assembly_method=None,
                                 doc=None):
        """Assembles ComponentDefinition into a linear primary structure.

        The resulting data structure is a partial design,
        still lacking an explicit sequence.
        To fully realize the target sequence, use Sequence::assemble().
        :param primary_structure: Either a list of URIs
        for the constituent ComponentDefinitions
        or a list of subcomponents.
        A list of displayIds is also acceptable if using SBOL-compliant URIs.
        :param assembly_standard: An optional argument
        such as IGEM_STANDARD_ASSEMBLY that affects
        how components are composed and the final target sequence.
        :param doc: The Document to which the assembled ComponentDefinitions
        will be added. If not set, then you must add this ComponentDefinition
        to a Document before calling this method.
        :return: None
        """
        primary_structure = self.assemble(primary_structure, assembly_method, doc)

        # If user specifies a list of IDs rather than ComponentDefinitions, convert to
        # list of ComponentDefinitions (Some parameter validation is done by the
        # preceding call to ComponentDefinition.assemble)
        doc = doc if doc else self.doc
        if all(isinstance(c, str) for c in primary_structure):
            component_identities = primary_structure[:]
            primary_structure = []
            for c_id in component_identities:
                cdef = doc.componentDefinitions[c_id]
                primary_structure.append(cdef)

        self.types += [SO_LINEAR]

        component_map = {}
        for c in self.components:
            if c.definition not in component_map:
                component_map[c.definition] = [c]
            else:
                component_map[c.definition].append(c)
        primary_structure_components = []
        for cd in primary_structure:
            primary_structure_components.append(component_map[cd.identity].pop())

        # Iterate pairwise through the primary_structure, and place SequenceConstraints
        # between adjacent ComponentDefinitions.
        if len(self.sequenceConstraints):
            self.sequenceConstraints.clear()
        for upstream, downstream in zip(primary_structure_components[:-1],
                                        primary_structure_components[1:]):
            instance_count = 0
            constraint_id = 'constraint_%d' % instance_count
            while constraint_id in self.sequenceConstraints:
                instance_count += 1
                constraint_id = 'constraint_%d' % instance_count
            sc = self.sequenceConstraints.create(constraint_id)
            sc.subject = upstream
            sc.object = downstream
            sc.restriction = SBOL_RESTRICTION_PRECEDES

    def compile(self, assembly_method=None):
        """Compiles an abstraction hierarchy of ComponentDefinitions
        into a nucleotide sequence.

        If no Sequence object is associated with this ComponentDefinition,
        one will be automatically instantiated
        :return: A string representing the nucleotide sequence
        for this ComponentDefinition.
        """
        if not self.doc:
            raise ValueError('Cannot compile <%s>. The ComponentDefinition must belong '
                             'to a Document in order to compile.' % self.identity)

        if self.sequence is None:
            sequence_id = self.displayId + '_seq'
            compliant_uris = Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS)
            typed_uris = Config.getOption(ConfigOptions.SBOL_TYPED_URIS)
            if compliant_uris and typed_uris:
                sequence_id = self.displayId
            self.sequence = Sequence(sequence_id)

        return self.sequence.compile(assembly_method=assembly_method)

    def updateSequence(self, composite_sequence=""):
        """Assemble a parent ComponentDefinition's Sequence
        from its subcomponent Sequences.

        :param composite_sequence: A recursive parameter, use default value.
        :return: The assembled parent sequence.
        """
        raise NotImplementedError("Not yet implemented")

    def getInSequentialOrder(self):
        """Get the primary sequence of a design in terms of its sequentially ordered
        Components.

        :return: A list of Components.
        """
        return self.getPrimaryStructureComponents()

    def hasUpstreamComponent(self, component):
        """Checks if the specified Component has a Component upstream
        in linear arrangement on the DNA strand.

        Checks that the appropriate SequenceConstraint exists.
        :param current_component: A Component in this ComponentDefinition.
        :return: True if found, False if not
        """
        if len(self.sequenceConstraints) < 1:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                            'Cannot determine upstream Component. '
                            'Self has no SequenceConstraints')
        else:
            for sc in self.sequenceConstraints:
                if sc.object == component.identity and \
                      sc.restriction == SBOL_RESTRICTION_PRECEDES:
                    return True
            return False

    def getUpstreamComponent(self, component):
        """Get the upstream component.

        :param component: A Component in this ComponentDefinition.
        :return: The upstream component.
        """
        if len(self.sequenceConstraints) < 1:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                            'Cannot get upstream Component. Self '
                            'has no SequenceConstraints')
        else:
            upstream_component_id = None
            for sc in self.sequenceConstraints:
                if sc.object == component.identity and \
                      sc.restriction == SBOL_RESTRICTION_PRECEDES:
                    upstream_component = self.components[sc.subject]
                    return upstream_component
        raise SBOLError(SBOLErrorCode.SBOL_ERROR_END_OF_LIST,
                        'This component has no upstream '
                        'component. Use hasUpstreamComponent to catch this error')

    def hasDownstreamComponent(self, component):
        """Checks if the specified Component has a Component downstream
        in linear arrangement on the DNA strand.

        Checks that the appropriate SequenceConstraint exists.
        :param current_component: A Component in this ComponentDefinition.
        :return: True if found, False if not.
        """
        if len(self.sequenceConstraints) < 1:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                            'Cannot determine upstream Component. '
                            'Self has no SequenceConstraints')
        else:
            for sc in self.sequenceConstraints:
                if sc.subject == component.identity and \
                      sc.restriction == SBOL_RESTRICTION_PRECEDES:
                    return True
            return False

    def getDownstreamComponent(self, component):
        """Get the downstream component.

        :param current_component: A Component in this ComponentDefinition.
        :return: The downstream component.
        """
        if len(self.sequenceConstraints) < 1:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                            'Cannot get downstream Component. '
                            'Self has no SequenceConstraints')
        else:
            upstream_component_id = None
            for sc in self.sequenceConstraints:
                if sc.subject == component.identity and \
                      sc.restriction == SBOL_RESTRICTION_PRECEDES:
                    upstream_component = self.components[sc.object]
                    return upstream_component
        raise SBOLError(SBOLErrorCode.SBOL_ERROR_END_OF_LIST,
                        'This component has no downstream '
                        'component. Use hasDownstreamComponent to catch this error')

    def deleteDownstreamComponent(self, upstream_component):
        if not Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS):
            raise ValueError('SBOL-compliant URIs must be enabled to use this method')
        if upstream_component.identity not in self.components:
            msg = 'Deletion failed. ComponentDefinition %s has no child component %s'
            msg = msg % (self.identity, upstream_component.identity)
            raise ValueError(msg)
        primary_structure = self.getPrimaryStructureComponents()
        if upstream_component.identity == primary_structure[-1].identity:
            msg = 'Deletion failed. No Components were found downstream of %s'
            msg = msg % upstream_component.identity
            raise ValueError(
                msg)
        downstream_component = None
        upstream_sequence_constraint = None
        downstream_sequence_constraint = None
        for c_upstream, c_downstream in zip(primary_structure[:-1],
                                            primary_structure[1:]):
            for sc in self.sequenceConstraints:
                if (sc.subject == c_upstream.identity and
                        sc.object == c_downstream.identity and
                        sc.restriction == SBOL_RESTRICTION_PRECEDES):
                    upstream_sequence_constraint = downstream_sequence_constraint
                    downstream_sequence_constraint = sc
            if downstream_component:
                break
            if c_upstream.identity == upstream_component.identity:
                downstream_component = c_downstream
        if downstream_component:
            self.components.remove(downstream_component.identity)
            self.sequenceConstraints.remove(downstream_sequence_constraint.identity)
            # The following condition is False when the downstream component is the last
            # component
            if downstream_sequence_constraint.subject == downstream_component.identity:
                upstream_sequence_constraint.object = \
                    downstream_sequence_constraint.object

    def deleteUpstreamComponent(self, downstream_component):
        if not Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS):
            raise ValueError('SBOL-compliant URIs must be enabled to use this method')
        if downstream_component.identity not in self.components:
            msg = 'Deletion failed. No Components were found upstream of %s'
            msg = msg % downstream_component.identity
            raise ValueError(msg)
        primary_structure = self.getPrimaryStructureComponents()
        if downstream_component.identity == primary_structure[0].identity:
            msg = 'Deletion failed. Component %s does not have an upstream component'
            msg = msg % downstream_component.identity
            raise ValueError(msg)
        upstream_component = None
        upstream_sequence_constraint = None
        downstream_sequence_constraint = None
        for c_upstream, c_downstream in zip(primary_structure[:-1],
                                            primary_structure[1:]):
            for sc in self.sequenceConstraints:
                if (sc.subject == c_upstream.identity and
                        sc.object == c_downstream.identity and
                        sc.restriction == SBOL_RESTRICTION_PRECEDES):
                    upstream_sequence_constraint = downstream_sequence_constraint
                    downstream_sequence_constraint = sc
            if c_downstream.identity == downstream_component.identity:
                upstream_component = c_upstream
                break
        if upstream_component:
            self.components.remove(upstream_component.identity)
            self.sequenceConstraints.remove(downstream_sequence_constraint.identity)
            if upstream_sequence_constraint:
                upstream_sequence_constraint.object = \
                    downstream_sequence_constraint.object

    def insertUpstreamComponent(self, downstream, insert):
        if not Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS):
            raise ValueError('SBOL-compliant URIs must be enabled to use this method')
        if not self.doc:
            msg = f'ComponentDefinition {self.identity} does not belong to a Document'
            msg += ' Add this ComponentDefinition to a Document before calling'
            msg += ' insertUpstreamComponent'
            raise ValueError(msg)
        if self.doc != insert.doc:
            msg = f'Invalid Document for ComponentDefinition {insert.identity}.'
            msg += ' Add the insert to the same Document as the calling object.'
            raise ValueError(msg)
        # If the user makes a mistake and tries to insert a ComponentDefinition
        # that doesn't already belong to this Document
        if not insert.doc:
            insert.doc = self.doc
        # Two cases. In first case, insert a Component that already has a
        # downstream Component specified by a SequenceConstraint. Otherwise,
        # append this Component to the end os sequential constraints.

        # Search for an existing SequenceConstraint between upstream
        # and downstream Component
        target_constraint = None
        for sc in self.sequenceConstraints:
            if (sc.object == downstream.identity and
                    sc.restriction == SBOL_RESTRICTION_PRECEDES):
                if target_constraint is not None:
                    # If more than one downstream component has been specified,
                    # then it is ambiguous where the insert should be placed,
                    # so throw an error
                    msg = 'SequenceConstraints are ambiguous. The target component'
                    msg += ' may have more than one downstream component specified'
                    raise ValueError(msg)
                target_constraint = sc
        # Generate URI of a Component to be created.  Check if an object with
        # that URI is already instantiated.
        instance_count = 0
        component_id = posixpath.join(self.persistentIdentity,
                                      f'{insert.displayId}_{instance_count}',
                                      self.version)
        while self.find(component_id):
            instance_count += 1
            component_id = posixpath.join(self.persistentIdentity,
                                          f'{insert.displayId}_{instance_count}',
                                          self.version)
        # Auto-construct the new Component
        c_insert = self.components.create(f'{insert.displayId}_{instance_count}')
        c_insert.definition = insert.identity
        # Generate URI of new SequenceConstraint. Check if an object with
        # that URI is already instantiated.
        instance_count = 0
        sc_id = posixpath.join(self.persistentIdentity,
                               f'constraint_{instance_count}',
                               self.version)
        while self.find(sc_id):
            instance_count += 1
            sc_id = posixpath.join(self.persistentIdentity,
                                   f'constraint_{instance_count}',
                                   self.version)
        # Auto-construct the new SequenceConstraint
        sc_new = self.sequenceConstraints.create(f'constraint_{instance_count}')
        sc_new.subject = component_id
        sc_new.object = downstream.identity
        sc_new.restriction = SBOL_RESTRICTION_PRECEDES
        # In case an upstream component was found...
        if target_constraint:
            target_constraint.object = c_insert.identity

    def insertDownstreamComponent(self, upstream, insert):
        if not Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS):
            raise ValueError('SBOL-compliant URIs must be enabled to use this method')
        if not self.doc:
            msg = f'ComponentDefinition {self.identity} does not belong to a Document'
            msg += ' Add this ComponentDefinition to a Document before calling'
            msg += ' insertUpstreamComponent'
            raise ValueError(msg)
        if self.doc != insert.doc:
            msg = f'Invalid Document for ComponentDefinition {insert.identity}.'
            msg += ' Add the insert to the same Document as the calling object.'
            raise ValueError(msg)
        # If the user makes a mistake and tries to insert a ComponentDefinition
        # that doesn't already belong to this Document
        if not insert.doc:
            insert.doc = self.doc
        # Two cases. In first case, insert a Component that already has a
        # downstream Component specified by a SequenceConstraint. Otherwise,
        # append this Component to the end os sequential constraints.

        # Search for an existing SequenceConstraint between upstream
        # and downstream Component
        target_constraint = None
        for sc in self.sequenceConstraints:
            if (sc.subject == upstream.identity and
                    sc.restriction == SBOL_RESTRICTION_PRECEDES):
                if target_constraint is not None:
                    # If more than one downstream component has been specified,
                    # then it is ambiguous where the insert should be placed,
                    # so throw an error
                    msg = 'SequenceConstraints are ambiguous. The target component'
                    msg += ' may have more than one downstream component specified'
                    raise ValueError(msg)
                target_constraint = sc
        # Generate URI of a Component to be created.  Check if an object with
        # that URI is already instantiated.
        instance_count = 0
        component_id = posixpath.join(self.persistentIdentity,
                                      f'{insert.displayId}_{instance_count}',
                                      self.version)
        while self.find(component_id):
            instance_count += 1
            component_id = posixpath.join(self.persistentIdentity,
                                          f'{insert.displayId}_{instance_count}',
                                          self.version)
        # Auto-construct the new Component
        c_insert = self.components.create(f'{insert.displayId}_{instance_count}')
        c_insert.definition = insert.identity
        # Generate URI of new SequenceConstraint. Check if an object with
        # that URI is already instantiated.
        instance_count = 0
        sc_id = posixpath.join(self.persistentIdentity,
                               f'constraint_{instance_count}',
                               self.version)
        while self.find(sc_id):
            instance_count += 1
            sc_id = posixpath.join(self.persistentIdentity,
                                   f'constraint_{instance_count}',
                                   self.version)
        # Auto-construct the new SequenceConstraint
        sc_new = self.sequenceConstraints.create(f'constraint_{instance_count}')
        sc_new.subject = upstream.identity
        sc_new.object = component_id
        sc_new.restriction = SBOL_RESTRICTION_PRECEDES
        # In case an upstream component was found...
        if target_constraint:
            target_constraint.object = c_insert.identity

    def getFirstComponent(self):
        """Gets the first Component in a linear sequence.

        :return: The first component in sequential order.
        """

        # A Component's sequential position in the primary structure does not
        # necessarily correspond with its index in the components list (Rather
        # this must be determined by reasoning over SequenceConstraints)
        if len(self.components) < 1:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                            'This ComponentDefinition has no '
                            'components')

        arbitrary_component = self.components[0]
        next_component = arbitrary_component
        while self.hasUpstreamComponent(next_component):
            next_component = self.getUpstreamComponent(next_component)
        return next_component

    def getLastComponent(self):
        """Gets the last Component in a linear sequence.

        :return: The last component in sequential order.
        """

        # A Component's sequential position in the primary structure does not
        # necessarily correspond with its index in the components list (Rather
        # this must be determined by reasoning over SequenceConstraints)
        if len(self.components) < 1:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                            'This ComponentDefinition has no '
                            'components')

        arbitrary_component = self.components[0]
        next_component = arbitrary_component
        while self.hasDownstreamComponent(next_component):
            next_component = self.getDownstreamComponent(next_component)
        return next_component

    def applyToComponentHierarchy(self, callback=None, user_data=None):
        """Perform an operation on every Component in a structurally-linked
        hierarchy of Components by applying a callback function.
        If no callback is specified, the default behavior is to return
        a list of each Component in the hierarchy.

        :param callback: The callback function to apply.
        :param user_data: Arbitrary user data which can be passed
        in and out of the callback as an argument.
        :return: A list of all Components in the hierarchy.
        """
        raise NotImplementedError("Not yet implemented")

    def getPrimaryStructureComponents(self):
        """Get the primary sequence of a design in terms of its sequentially ordered
        Components.

        :return: A list of Components.
        """
        subcomponents = []
        if len(self.components) == 1:
            subcomponents.append(self.components[0])
        else:
            # Check if this is a complete primary structure
            # (note this isn't a perfect test)
            if len(self.sequenceConstraints) != (len(self.components) - 1):
                raise ValueError('ComponentDefinition <%s> does not appear to describe'
                                 'a complete primary structure. It appears to be '
                                 'missing SequenceConstraints.' % self.identity)

            c_first = self.getFirstComponent()
            subcomponents.append(c_first)
            c_next = c_first
            while self.hasDownstreamComponent(c_next):
                c_next = self.getDownstreamComponent(c_next)
                subcomponents.append(c_next)
        return subcomponents

    def getPrimaryStructure(self):
        """Get the primary sequence of a design in terms of its sequentially ordered
        ComponentDefinitions.

        :return: A list of ComponentDefinitions.
        """
        if self.doc is None:
            raise SBOLError(SBOLErrorCode.SBOL_ERROR_MISSING_DOCUMENT,
                            'Cannot get primary structure.'
                            'Self must belong to a Document.')
        component_ids = [c.definition for c in self.getPrimaryStructureComponents()]
        return [self.doc.getComponentDefinition(c) for c in component_ids]

    def insertDownstream(self, target, component_to_insert):
        """Insert a Component downstream of another in a primary sequence,
        shifting any adjacent Components downstream as well.

        :param target: The target Component will be upstream
        of the insert Component after this operation.
        :param component_to_insert: The insert Component
        is inserted downstream of the target Component.
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def insertUpstream(self, target, component_to_insert):
        """Insert a Component upstream of another in a primary sequence,
        shifting any adjacent Components upstream as well.


        :param target: The target Component will be downstream
        of the insert Component after this operation.
        :param component_to_insert: The insert Component is inserted upstream
        of the target Component.
        :return:
        """
        raise NotImplementedError("Not yet implemented")

    def addUpstreamFlank(self, target, elements):
        """A useful method when building up SBOL representations
        of natural DNA sequences.

        For example, it is often necessary to specify components
        that are assumed to have no meaningful role in the design,
        but are nevertheless important to fill in regions of sequence.
        This method autoconstructs a ComponentDefinition and Sequence object
        to create an arbitrary flanking sequence around design Components.
        The new ComponentDefinition will have Sequence Ontology type
        of flanking_region or SO:0000239.
        :param target: The new flanking sequence will be placed upstream
        of the target.
        :param elements: The primary sequence elements will be assigned
        to the autoconstructed Sequence object.
        The encoding is inferred.
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def addDownstreamFlank(self, target, elements):
        """A useful method when building up SBOL representations
        of natural DNA sequences.

        For example, it is often necessary to specify components
        that are assumed to have no meaningful role
        in the design, but are nevertheless important
        to fill in regions of sequence. This method autoconstructs
        a ComponentDefinition and Sequence object to create an arbitrary
        flanking sequence around design Components.
        The new ComponentDefinition will have Sequence Ontology type
        of flanking_sequence.
        :param target: The new flanking sequence will be placed downstream
        of the target.
        :param elements: The primary sequence elements will be assigned
        to the autoconstructed Sequence object.
        The encoding is inferred.
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def isRegular(self, msg=None):
        """Use this diagnose an irregular design.

        Recursively checks if this ComponentDefinition
        defines a SequenceAnnotation and Range for every Sequence.
        Regularity is more stringent than completeness.
        A design must be complete to be regular.
        :param msg: An optional message for diagnosing the irregularity,
        if any is found.
        :return: True if the abstraction hierarchy is regular,
        False otherwise.
        """
        raise NotImplementedError("Not yet implemented")

    def isComplete(self, msg=None):
        """Use this diagnose an incomplete design.

        Recursively checks if this ComponentDefinition defines
        a SequenceAnnotation and Range for every Sequence.
        Completeness does not guarantee regularity.
        :param msg: An optional message for diagnosing the irregularity,
        if any is found.
        :return: True if the abstraction hierarchy is complete, False otherwise.
        """
        raise NotImplementedError("Not yet implemented")

    def disassemble(self, range_start=1):
        """Instantiates a Component for every SequenceAnnotation

         When converting from a flat GenBank file to a flat SBOL file,
         the result is a ComponentDefinition with SequenceAnnotations.
         This method will convert the flat SBOL file
         into hierarchical SBOL.

        :param range_start:
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def linearize(self, components=None):
        """
        TODO document

        :param components: An optional list of component definitions or URIs.
        If None, an empty list of ComponentDefinitions is assumed.
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def build(self):
        """
        TODO document

        :return: A ComponentDefinition.
        """
        raise NotImplementedError("Not yet implemented")

    def participate(self, species):
        """A convenience method that assigns a component to participate
        in a biochemical reaction.

        Behind the scenes, it auto-constructs a FunctionalComponent
        for this ComponentDefinition and assigns it to a Participation.

        :param species: A Participation object
        (ie. participant species in a biochemical Interaction).
        :return: None
        """
        raise NotImplementedError("Not yet implemented")

    def getTypeURI(self):
        return SBOL_COMPONENT_DEFINITION

    def integrateAtBaseCoordinate(self, target_cd, insert_cd, base_coordinate):
        """
        Construct SBOL representing a genetic insert. Inserts insert_cd
        into self at base_coordinate.

        This method constructs a new ComponentDefinition that is annotated
        with the original sequence and the inserted sequence such that the
        new DNA sequence can be generated. This method does not generate
        the new sequence itself.

        The new sequence is not generated to avoid duplicating very long
        sequences in memory when they are not needed.

        """

        def autoconstruct_id(sbol_owned_object_property, display_id):
            instance_count = 0
            auto_id = '%s_%d' % (display_id, instance_count)
            while sbol_owned_object_property.find(auto_id):
                instance_count += 1
                auto_id = '%s_%d' % (display_id, instance_count)
            return auto_id

        if not self.doc:
            msg = 'Integration failed.'
            msg += ' ComponentDefinition <%s> must be added to a Document'
            msg += ' in order to proceed with integration.'
            raise ValueError(msg % self.identity)
        if not target_cd.doc or target_cd.doc.this != self.doc.this:
            msg = 'Integration failed.'
            msg += ' The target_cd <%s> must be added to the same Document'
            msg += ' as self before proceeding.'
            raise ValueError(msg % target_cd.identity)
        if not insert_cd.doc or insert_cd.doc.this != self.doc.this:
            msg = 'Integration failed.'
            msg += ' The insert_cd <%s> must be added to the same Document'
            msg += ' as self before proceeding.'
            raise ValueError(msg % insert_cd.identity)
        if not target_cd.sequence:
            msg = 'Integration failed.'
            msg += ' The target_cd <%s> is not associated with a Sequence.'
            msg += ' The sequence property should point to a valid Sequence'
            msg += ' before proceeding.'
            raise ValueError(msg % self.identity)
        if not target_cd.sequence.elements:
            msg = 'Integration failed.'
            msg += ' The elements property of Sequence <%s> must be set'
            msg += ' before proceeding. The sequence property should point'
            msg += ' to a valid Sequence before proceeding.'
            raise ValueError(msg % self.sequence.identity)
        if not insert_cd.sequence:
            msg = 'Integration failed.'
            msg += ' The insert_cd <%s> is not associated with a Sequence.'
            msg += ' The sequence property must point to a valid Sequence'
            msg += ' before proceeding.'
            raise ValueError(msg % insert_cd.identity)
        if not insert_cd.sequence.elements:
            msg = 'Integration failed.'
            msg += ' The elements property of Sequence <%s> must be set'
            msg += ' before proceeding. The sequence property should point'
            msg += ' to a valid Sequence before proceeding.'
            raise ValueError(msg % insert_cd.sequence.identity)

        target_cd_comp = None
        insert_cd_comp = None
        for c in self.components:
            if c.definition == target_cd.identity:
                if not target_cd_comp:
                    target_cd_comp = c
                else:
                    msg = 'Integration failed.'
                    msg += ' Self contains more than one instance of %s'
                    raise ValueError(msg % target_cd.identity)
            if c.definition == insert_cd.identity:
                if not insert_cd_comp:
                    insert_cd_comp = c
                else:
                    msg = 'Integration failed.'
                    msg += ' Self contains more than one instance of %s'
                    raise ValueError(msg % insert_cd.identity)

        orig_len = len(target_cd.sequence.elements)
        insert_len = len(insert_cd.sequence.elements)

        # Keep base_coordinate in bounds
        if base_coordinate < 1:
            msg = 'Insert failed. The base_coordinate must be a base'
            msg += ' coordinate equal to or greater than 1'
            raise ValueError(msg)
        if base_coordinate > orig_len + 1:
            msg = 'Insert failed. The base_coordinate exceeds the'
            msg += ' length of the target sequence.'
            raise ValueError(msg)

        target_cd_comp = None
        if base_coordinate > 1:
            # Now link target_cd into the structure of the new
            # ComponentDefinition
            if not target_cd_comp:
                new_id = autoconstruct_id(self.components, target_cd.displayId)
                target_cd_comp = self.components.create(new_id)
                target_cd_comp.definition = target_cd
            new_id = autoconstruct_id(target_cd_comp.sourceLocations,
                                      target_cd.displayId)
            source_loc = target_cd_comp.sourceLocations.createRange(new_id)
            source_loc.start = 1
            source_loc.end = base_coordinate - 1

        # Now link the insert to the new cd
        if not insert_cd_comp:
            new_id = autoconstruct_id(self.components, insert_cd.displayId)
            insert_cd_comp = self.components.create(new_id)
            insert_cd_comp.definition = insert_cd

        target_cd_comp_1 = None
        if base_coordinate <= orig_len:
            new_id = autoconstruct_id(self.components, target_cd.displayId)
            target_cd_comp_1 = self.components.create(new_id)
            target_cd_comp_1.definition = target_cd
            new_id = autoconstruct_id(target_cd_comp_1.sourceLocations,
                                      target_cd.displayId)
            source_loc_1 = target_cd_comp_1.sourceLocations.createRange(new_id)
            source_loc_1.start = base_coordinate
            source_loc_1.end = orig_len

        if target_cd_comp:
            new_id = autoconstruct_id(self.sequenceConstraints, self.displayId)
            sc0 = self.sequenceConstraints.create(new_id)
            sc0.subject = target_cd_comp
            sc0.object = insert_cd_comp
            sc0.restriction = SBOL_RESTRICTION_PRECEDES

        if target_cd_comp_1:
            new_id = autoconstruct_id(self.sequenceConstraints, self.displayId)
            sc1 = self.sequenceConstraints.create(new_id)
            sc1.subject = insert_cd_comp
            sc1.object = target_cd_comp_1
            sc1.restriction = SBOL_RESTRICTION_PRECEDES



import unittest

import sbol2


class TestVersionProperty(unittest.TestCase):

    def test_values(self):
        cd = sbol2.ComponentDefinition('cd1')
        self.assertEqual('1', cd.version)
        with self.assertRaises(ValueError):
            cd.version = 'tom'
        with self.assertRaises(ValueError):
            # Floats are not acceptable, must be a string
            cd.version = 1.2
        v = '2'
        cd.version = v
        self.assertEqual(v, cd.version)
        v = '2.3'
        cd.version = v
        self.assertEqual(v, cd.version)
        v = '2.3.1'
        cd.version = v
        self.assertEqual(v, cd.version)

    def test_init_store(self):
        # Ensure that property constructors initialize the parent
        # object's value store
        obj = sbol2.SBOLObject()
        type_uri = 'http://example.com#thing'
        obj.thing = sbol2.VersionProperty(obj, type_uri, '0', '*')
        self.assertIn(type_uri, obj.properties)
        self.assertEqual([], obj.properties[type_uri])
        self.assertEqual([], obj.thing)


if __name__ == '__main__':
    unittest.main()



from .identified import Identified
from .constants import *
from rdflib import URIRef
from .property import URIProperty, ReferencedObject


class MapsTo(Identified):

    def __init__(self, type_uri=SBOL_MAPS_TO, uri=URIRef('example'), local='',
                 remote='', refinement=SBOL_REFINEMENT_VERIFY_IDENTICAL,
                 version=VERSION_STRING):
        super().__init__(type_uri, uri, version)
        self.local = ReferencedObject(self, SBOL_LOCAL, SBOL_COMPONENT,
                                      '1', '1', [], local)
        self.remote = ReferencedObject(self, SBOL_REMOTE, SBOL_COMPONENT,
                                       '1', '1', [], remote)
        self.refinement = URIProperty(self, SBOL_REFINEMENT,
                                      '1', '1', [], refinement)



from enum import Enum
from typing import Any


class SBOLErrorCode(Enum):
    DUPLICATE_URI_ERROR = 1
    NOT_FOUND_ERROR = 2
    SBOL_ERROR_NOT_FOUND = 3
    END_OF_LIST = 4
    SBOL_ERROR_END_OF_LIST = 5
    SBOL_ERROR_SERIALIZATION = 6
    SBOL_ERROR_PARSE = 7
    SBOL_ERROR_MISSING_NAMESPACE = 8
    SBOL_ERROR_NONCOMPLIANT_VERSION = 9
    SBOL_ERROR_COMPLIANCE = 10
    SBOL_ERROR_MISSING_DOCUMENT = 11
    SBOL_ERROR_INVALID_ARGUMENT = 12
    SBOL_ERROR_FILE_NOT_FOUND = 13
    SBOL_ERROR_ORPHAN_OBJECT = 14
    SBOL_ERROR_TYPE_MISMATCH = 15
    SBOL_ERROR_BAD_HTTP_REQUEST = 16
    SBOL_ERROR_URI_NOT_UNIQUE = 17
    SBOL_ERROR_HTTP_UNAUTHORIZED = 18


class SBOLError(Exception):

    def __init__(self, err: SBOLErrorCode, message: Any):
        if not isinstance(err, SBOLErrorCode):
            raise TypeError('SBOLError err argument must be a SBOLErrorCode')
        self._message = str(message)
        self._err = err

    def what(self):
        return self._message

    def error_code(self):
        return self._err



import unittest
import sbol2


class TestParticipation(unittest.TestCase):

    def test_add_remove_role(self):
        p = sbol2.Participation('p')
        self.assertEqual([], p.roles)
        p.addRole(sbol2.SO_PROMOTER)
        self.assertEqual([sbol2.SO_PROMOTER], p.roles)
        p.addRole(sbol2.SO_MISC)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC], p.roles)
        p.addRole(sbol2.SO_CDS)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_MISC, sbol2.SO_CDS], p.roles)
        p.removeRole(1)
        self.assertEqual([sbol2.SO_PROMOTER, sbol2.SO_CDS], p.roles)


if __name__ == '__main__':
    unittest.main()



from .config import parseNamespace, parseClassName
from .constants import *
from .property import URIProperty, IntProperty, TextProperty
from .toplevel import TopLevel

SEARCH_QUERY_URI: str = PYSBOL2_NS.SearchQuery.toPython()
OBJECT_TYPE_URI: str = PYSBOL2_NS.ObjectType.toPython()
OFFSET_URI: str = PYSBOL2_NS.Offset.toPython()
LIMIT_URI: str = PYSBOL2_NS.Limit.toPython()


class SearchQuery(TopLevel):
    """Search terms to support PartShop advanced search."""
    def __init__(self, search_target: str = SBOL_COMPONENT_DEFINITION,
                 offset: int = 0, limit: int = 25):
        super().__init__(SEARCH_QUERY_URI)
        # Set attributes from TopLevel
        self.displayId = ''
        self.persistentIdentity = ''
        self.version = ''
        # Add some attributes
        self.objectType = URIProperty(self, OBJECT_TYPE_URI, '0', '1', None,
                                      search_target)
        self.offset = IntProperty(self, OFFSET_URI, '0', '1', None, offset)
        self.limit = IntProperty(self, LIMIT_URI, '0', '1', None, limit)

    def __setitem__(self, key, value):
        if parseNamespace(key) == '':
            uri = SBOL_URI + '#' + 'key'
        else:
            uri = key
        TextProperty(self, uri, '0', '1', None, value)

    def __getitem__(self, item):
        return self.properties[item][0].toPython()

    def query_dict(self) -> dict:
        """Convert this search query to a query dictionary suitable for
        urllib.parse.urlencode()."""
        result = {}
        object_type = self.objectType
        if object_type:
            result['objectType'] = parseClassName(object_type)
        collection = self.getPropertyValue(SBOL_COLLECTION)
        if collection:
            result['collection'] = f'<{collection}>'
        # Now gather all the properties the user has set
        skip_list = [SBOL_IDENTITY,
                     OBJECT_TYPE_URI, OFFSET_URI, LIMIT_URI,
                     SBOL_COLLECTION]
        for k, v in self.properties.items():
            if k in skip_list:
                continue
            if not v or not v[0]:
                # If there is no value, skip it
                # this can happen if no value is supplied for an attribute
                continue
            if v[0].startswith('http'):
                result[f'<{k}>'] = f'<{v[0]}>'
            else:
                result[f'<{k}>'] = f"'{v[0]}'"
        return result



from sbol2 import *

setHomespace('http://sbols.org/CRISPR_Example')
Config.setOption(ConfigOptions.SBOL_TYPED_URIS, False)
version = '1.0'
doc = Document()

cas9_generic = ComponentDefinition('cas9_generic', BIOPAX_PROTEIN, version)
doc.addComponentDefinition(cas9_generic)

gRNA_generic = ComponentDefinition('gRNA_generic', BIOPAX_RNA, version)
gRNA_generic.roles = [SO_SGRNA]
doc.addComponentDefinition(gRNA_generic)

cas9_gRNA_complex = ComponentDefinition('cas9_gRNA_complex', BIOPAX_COMPLEX, version)
doc.addComponentDefinition(cas9_gRNA_complex)

target_gene = ComponentDefinition('target_gene', BIOPAX_DNA, version)
target_gene.roles = [SO_PROMOTER]
doc.addComponentDefinition(target_gene)

target = ComponentDefinition('target', BIOPAX_PROTEIN, version)
doc.addComponentDefinition(target)

CRISPR_Template = ModuleDefinition('CRISPR_Template', version)
doc.addModuleDefinition(CRISPR_Template)

Cas9Complex_Formation = CRISPR_Template.interactions.create('cas9_complex_formation')
Cas9Complex_Formation.types = [SBO_NONCOVALENT_BINDING]

cas9_generic_fc = CRISPR_Template.functionalComponents.create('cas9_generic')
cas9_generic_fc.definition = cas9_generic.persistentIdentity
cas9_generic_fc.access = SBOL_ACCESS_PUBLIC
cas9_generic_fc.direction = SBOL_DIRECTION_IN_OUT
cas9_generic_fc.version = version

cas9_generic_participation = Cas9Complex_Formation.participations.create('cas9_generic')
cas9_generic_participation.roles = [SBO_REACTANT]
cas9_generic_participation.participant = cas9_generic_fc.identity

gRNA_generic_fc = CRISPR_Template.functionalComponents.create('gRNA_generic')
gRNA_generic_fc.definition = gRNA_generic.persistentIdentity
gRNA_generic_fc.access = SBOL_ACCESS_PUBLIC
gRNA_generic_fc.direction = SBOL_DIRECTION_IN_OUT
gRNA_generic_fc.version = version

gRNA_generic_participation = Cas9Complex_Formation.participations.create('gRNA_generic')
gRNA_generic_participation.roles = [SBO_REACTANT]
gRNA_generic_participation.participant = gRNA_generic_fc.identity

cas9_gRNA_complex_fc = CRISPR_Template.functionalComponents.create('cas9_gRNA_complex')
cas9_gRNA_complex_fc.definition = cas9_gRNA_complex.persistentIdentity
cas9_gRNA_complex_fc.access = SBOL_ACCESS_PUBLIC
cas9_gRNA_complex_fc.direction = SBOL_DIRECTION_IN_OUT
cas9_gRNA_complex_fc.version = version

cas9_gRNA_complex_participation = \
    Cas9Complex_Formation.participations.create('cas9_gRNA_complex')
cas9_gRNA_complex_participation.roles = [SBO_PRODUCT]
cas9_gRNA_complex_participation.participant = cas9_gRNA_complex_fc.identity

EYFP_production = CRISPR_Template.interactions.create('target_production')
EYFP_production.types = [SBO_GENETIC_PRODUCTION]

target_gene_fc = CRISPR_Template.functionalComponents.create('target_gene')
target_gene_fc.definition = target_gene.persistentIdentity
target_gene_fc.access = SBOL_ACCESS_PUBLIC
target_gene_fc.direction = SBOL_DIRECTION_IN_OUT
target_gene_fc.version = version

target_gene_participation = EYFP_production.participations.create('target_gene')
target_gene_participation.roles = [SBO_PROMOTER]
target_gene_participation.participant = target_gene_fc.identity

target_fc = CRISPR_Template.functionalComponents.create('target')
target_fc.definition = target.persistentIdentity
target_fc.access = SBOL_ACCESS_PUBLIC
target_fc.direction = SBOL_DIRECTION_IN_OUT
target_fc.version = version

target_participation = EYFP_production.participations.create('target')
target_participation.roles = [SBO_PRODUCT]
target_participation.participant = target_fc.identity

target_generic_gene_inhibition = \
    CRISPR_Template.interactions.create('target_gene_inhibition')
target_generic_gene_inhibition.types = [SBO_INHIBITION]

cas9_gRNA_complex_participation1 = \
    target_generic_gene_inhibition.participations.create('cas9_gRNA_complex')
cas9_gRNA_complex_participation1.roles = [SBO_INHIBITOR]
cas9_gRNA_complex_participation1.participant = cas9_gRNA_complex_fc.identity

target_gene_participation2 = \
    target_generic_gene_inhibition.participations.create('target_gene')
target_gene_participation2.roles = [SBO_PROMOTER]
target_gene_participation2.participant = target_gene_fc.identity

CRa_U6_seq_elements = ('GGTTTACCGAGCTCTTATTGGTTTTCAAACTTCATTGACTGTGCC'
                       'AAGGTCGGGCAGGAAGAGGGCCTATTTCCCATGATTCCTTCATAT'
                       'TTGCATATACGATACAAGGCTGTTAGAGAGATAATTAGAATTAAT'
                       'TTGACTGTAAACACAAAGATATTAGTACAAAATACGTGACGTAGA'
                       'AAGTAATAATTTCTTGGGTAGTTTGCAGTTTTAAAATTATGTTTT'
                       'AAAATGGACTATCATATGCTTACCGTAACTTGAAATATAGAACCG'
                       'ATCCTCCCATTGGTATATATTATAGAACCGATCCTCCCATTGGCT'
                       'TGTGGAAAGGACGAAACACCGTACCTCATCAGGAACATGTGTTTA'
                       'AGAGCTATGCTGGAAACAGCAGAAATAGCAAGTTTAAATAAGGCT'
                       'AGTCCGTTATCAACTTGAAAAAGTGGCACCGAGTCGGTGCTTTTT'
                       'TTGGTGCGTTTTTATGCTTGTAGTATTGTATAATGTTTTT')
CRa_U6_seq = Sequence('CRa_U6_seq', CRa_U6_seq_elements, SBOL_ENCODING_IUPAC, version)
doc.addSequence(CRa_U6_seq)

gRNA_b_elements = ('AAGGTCGGGCAGGAAGAGGGCCTATTTCCCATGATTCCTTCATAT'
                   'TTGCATATACGATACAAGGCTGTTAGAGAGATAATTAGAATTAAT'
                   'TTGACTGTAAACACAAAGATATTAGTACAAAATACGTGACGTAGA'
                   'AAGTAATAATTTCTTGGGTAGTTTGCAGTTTTAAAATTATGTTTT'
                   'AAAATGGACTATCATATGCTTACCGTAACTTGAAAGTATTTCGAT'
                   'TTCTTGGCTTTATATATCTTGTGGAAAGGACGAAACACCGTACCT'
                   'CATCAGGAACATGTGTTTAAGAGCTATGCTGGAAACAGCAGAAAT'
                   'AGCAAGTTTAAATAAGGCTAGTCCGTTATCAACTTGAAAAAGTGG'
                   'CACCGAGTCGGTGCTTTTTTT')
gRNA_b_seq = Sequence('gRNA_b_seq', gRNA_b_elements, SBOL_ENCODING_IUPAC, version)
doc.addSequence(gRNA_b_seq)

mKate_seq_elements = ('TCTAAGGGCGAAGAGCTGATTAAGGAGAACATGCACATGAAGCTG'
                      'TACATGGAGGGCACCGTGAACAACCACCACTTCAAGTGCACATCC'
                      'GAGGGCGAAGGCAAGCCCTACGAGGGCACCCAGACCATGAGAATC'
                      'AAGGTGGTCGAGGGCGGCCCTCTCCCCTTCGCCTTCGACATCCTG'
                      'GCTACCAGCTTCATGTACGGCAGCAAAACCTTCATCAACCACACC'
                      'CAGGGCATCCCCGACTTCTTTAAGCAGTCCTTCCCTGAGGTAAGT'
                      'GGTCCTACCTCATCAGGAACATGTGTTTTAGAGCTAGAAATAGCA'
                      'AGTTAAAATAAGGCTAGTCCGTTATCAACTTGAAAAAGTGGCACC'
                      'GAGTCGGTGCTACTAACTCTCGAGTCTTCTTTTTTTTTTTCACAG'
                      'GGCTTCACATGGGAGAGAGTCACCACATACGAAGACGGGGGCGTG'
                      'CTGACCGCTACCCAGGACACCAGCCTCCAGGACGGCTGCCTCATC'
                      'TACAACGTCAAGATCAGAGGGGTGAACTTCCCATCCAACGGCCCT'
                      'GTGATGCAGAAGAAAACACTCGGCTGGGAGGCCTCCACCGAGATG'
                      'CTGTACCCCGCTGACGGCGGCCTGGAAGGCAGAAGCGACATGGCC'
                      'CTGAAGCTCGTGGGCGGGGGCCACCTGATCTGCAACTTGAAGACC'
                      'ACATACAGATCCAAGAAACCCGCTAAGAACCTCAAGATGCCCGGC'
                      'GTCTACTATGTGGACAGAAGACTGGAAAGAATCAAGGAGGCCGAC'
                      'AAAGAGACCTACGTCGAGCAGCACGAGGTGGCTGTGGCCAGATAC'
                      'TGCG')
mKate_seq = Sequence('mKate_seq', mKate_seq_elements, SBOL_ENCODING_IUPAC, version)
doc.addSequence(mKate_seq)

CRP_b_seq_elements = ('GCTCCGAATTTCTCGACAGATCTCATGTGATTACGCCAAGCTACG'
                      'GGCGGAGTACTGTCCTCCGAGCGGAGTACTGTCCTCCGAGCGGAG'
                      'TACTGTCCTCCGAGCGGAGTACTGTCCTCCGAGCGGAGTTCTGTC'
                      'CTCCGAGCGGAGACTCTAGATACCTCATCAGGAACATGTTGGAAT'
                      'TCTAGGCGTGTACGGTGGGAGGCCTATATAAGCAGAGCTCGTTTA'
                      'GTGAACCGTCAGATCGCCTCGAGTACCTCATCAGGAACATGTTGG'
                      'ATCCAATTCGACC')
CRP_b_seq = Sequence('CRP_b_seq', CRP_b_seq_elements, SBOL_ENCODING_IUPAC, version)
doc.addSequence(CRP_b_seq)

pConst = ComponentDefinition('pConst', BIOPAX_DNA, version)
pConst.roles = [SO_PROMOTER]
doc.addComponentDefinition(pConst)

cas9m_BFP_cds = ComponentDefinition('cas9m_BFP_cds', BIOPAX_DNA, version)
cas9m_BFP_cds.roles = [SO_CDS]
doc.addComponentDefinition(cas9m_BFP_cds)

cas9m_BFP_gene = ComponentDefinition('cas9m_BFP_gene', BIOPAX_DNA, version)
cas9m_BFP_gene.roles = [SO_PROMOTER]
doc.addComponentDefinition(cas9m_BFP_gene)

pConst_c2 = cas9m_BFP_gene.components.create('pConst')
pConst_c2.definition = pConst.persistentIdentity
pConst_c2.access = SBOL_ACCESS_PUBLIC
pConst_c2.version = version

cas9m_BFP_cds_c = cas9m_BFP_gene.components.create('cas9m_BFP_cds')
cas9m_BFP_cds_c.definition = cas9m_BFP_cds.persistentIdentity
cas9m_BFP_cds_c.access = SBOL_ACCESS_PUBLIC
cas9m_BFP_cds_c.version = version

cas9m_BFP_gene_constraint = \
    cas9m_BFP_gene.sequenceConstraints.create('cas9m_BFP_gene_constraint')
cas9m_BFP_gene_constraint.subject = pConst_c2.identity
cas9m_BFP_gene_constraint.object = cas9m_BFP_cds_c.identity
cas9m_BFP_gene_constraint.restriction = SBOL_RESTRICTION_PRECEDES

cas9m_BFP = ComponentDefinition('cas9m_BFP', BIOPAX_PROTEIN, version)
doc.addComponentDefinition(cas9m_BFP)

CRa_U6 = ComponentDefinition('CRa_U6', BIOPAX_DNA, version)
CRa_U6.sequences = [CRa_U6_seq.persistentIdentity]
CRa_U6.roles = [SO_PROMOTER]
doc.addComponentDefinition(CRa_U6)

gRNA_b_nc = ComponentDefinition('gRNA_b_nc', BIOPAX_DNA, version)
gRNA_b_nc.roles = [SO_CDS]
gRNA_b_nc.sequences = [gRNA_b_seq.persistentIdentity]
doc.addComponentDefinition(gRNA_b_nc)

gRNA_b_terminator = ComponentDefinition('gRNA_b_terminator', BIOPAX_DNA, version)
gRNA_b_terminator.roles = [SO_TERMINATOR]
doc.addComponentDefinition(gRNA_b_terminator)

gRNA_b_gene = ComponentDefinition('gRNA_b_gene', BIOPAX_DNA, version)
gRNA_b_gene.roles = [SO_PROMOTER]
doc.addComponentDefinition(gRNA_b_gene)

CRa_U6_c = gRNA_b_gene.components.create('CRa_U6')
CRa_U6_c.definition = CRa_U6.persistentIdentity
CRa_U6_c.access = SBOL_ACCESS_PUBLIC
CRa_U6_c.version = version

gRNA_b_nc_c = gRNA_b_gene.components.create('gRNA_b_nc')
gRNA_b_nc_c.definition = gRNA_b_nc.persistentIdentity
gRNA_b_nc_c.access = SBOL_ACCESS_PUBLIC
gRNA_b_nc_c.version = version

gRNA_b_terminator_c = gRNA_b_gene.components.create('gRNA_b_terminator')
gRNA_b_terminator_c.definition = gRNA_b_terminator.persistentIdentity
gRNA_b_terminator_c.access = SBOL_ACCESS_PUBLIC
gRNA_b_terminator_c.version = version

gRNA_gene_constraint1 = gRNA_b_gene.sequenceConstraints.create('gRNA_b_gene_constraint1')
gRNA_gene_constraint1.subject = CRa_U6_c.identity
gRNA_gene_constraint1.object = gRNA_b_nc_c.identity
gRNA_gene_constraint1.restriction = SBOL_RESTRICTION_PRECEDES

gRNA_gene_constraint2 = gRNA_b_gene.sequenceConstraints.create('gRNA_b_gene_constraint2')
gRNA_gene_constraint2.subject = gRNA_b_nc_c.identity
gRNA_gene_constraint2.object = gRNA_b_terminator_c.identity
gRNA_gene_constraint2.restriction = SBOL_RESTRICTION_PRECEDES

gRNA_b = ComponentDefinition('gRNA_b', BIOPAX_RNA, version)
gRNA_b.roles = [SO_SGRNA]
doc.addComponentDefinition(gRNA_b)

cas9m_BFP_gRNA_b = ComponentDefinition('cas9m_BFP_gRNA_b', BIOPAX_COMPLEX, version)
doc.addComponentDefinition(cas9m_BFP_gRNA_b)

mKate_cds = ComponentDefinition('mKate_cds', BIOPAX_DNA, version)
mKate_cds.roles = [SO_CDS]
mKate_cds.sequences = [mKate_seq.persistentIdentity]
doc.addComponentDefinition(mKate_cds)

mKate_gene = ComponentDefinition('mKate_gene', BIOPAX_DNA, version)
mKate_gene.roles = [SO_PROMOTER]
doc.addComponentDefinition(mKate_gene)

pConst_c = mKate_gene.components.create('pConst')
pConst_c.definition = pConst.persistentIdentity
pConst_c.access = SBOL_ACCESS_PUBLIC
pConst_c.version = version

mKate_cds_c = mKate_gene.components.create('mKate_cds')
mKate_cds_c.definition = mKate_cds.persistentIdentity
mKate_cds_c.access = SBOL_ACCESS_PUBLIC
mKate_cds_c.version = version

mKate_gene_constraint = mKate_gene.sequenceConstraints.create('mKate_gene_constraint')
mKate_gene_constraint.subject = pConst_c.identity
mKate_gene_constraint.object = mKate_cds_c.identity
mKate_gene_constraint.restriction = SBOL_RESTRICTION_PRECEDES

mKate = ComponentDefinition('mKate', BIOPAX_PROTEIN, version)
doc.addComponentDefinition(mKate)

Gal4VP16_cds = ComponentDefinition('Gal4VP16_cds', BIOPAX_DNA, version)
Gal4VP16_cds.roles = [SO_CDS]
doc.addComponentDefinition(Gal4VP16_cds)

Gal4VP16_gene = ComponentDefinition('Gal4VP16_gene', BIOPAX_DNA, version)
Gal4VP16_gene.roles = [SO_PROMOTER]
doc.addComponentDefinition(Gal4VP16_gene)

pConst_c3 = Gal4VP16_gene.components.create('pConst')
pConst_c3.definition = pConst.persistentIdentity
pConst_c3.access = SBOL_ACCESS_PUBLIC
pConst_c3.version = version

Gal4VP16_cds_c = Gal4VP16_gene.components.create('Gal4VP16_cds')
Gal4VP16_cds_c.definition = Gal4VP16_cds.persistentIdentity
Gal4VP16_cds_c.access = SBOL_ACCESS_PUBLIC
Gal4VP16_cds_c.version = version

GAL4VP16_gene_constraint = \
    Gal4VP16_gene.sequenceConstraints.create('GAL4VP16_gene_constraint')
GAL4VP16_gene_constraint.subject = pConst_c3.identity
GAL4VP16_gene_constraint.object = Gal4VP16_cds_c.identity
GAL4VP16_gene_constraint.restriction = SBOL_RESTRICTION_PRECEDES

Gal4VP16 = ComponentDefinition('Gal4VP16', BIOPAX_PROTEIN, version)
doc.addComponentDefinition(Gal4VP16)

CRP_b = ComponentDefinition('CRP_b', BIOPAX_DNA, version)
CRP_b.roles = [SO_PROMOTER]
CRP_b.sequences = [CRP_b_seq.persistentIdentity]
doc.addComponentDefinition(CRP_b)

EYFP_cds = ComponentDefinition('EYFP_cds', BIOPAX_DNA, version)
EYFP_cds.roles = [SO_CDS]
doc.addComponentDefinition(EYFP_cds)

EYFP_gene = ComponentDefinition('EYFP_gene', BIOPAX_DNA, version)
EYFP_gene.roles = [SO_PROMOTER]
doc.addComponentDefinition(EYFP_gene)

CRP_b_c = EYFP_gene.components.create('CRP_b')
CRP_b_c.definition = CRP_b.persistentIdentity
CRP_b_c.access = SBOL_ACCESS_PUBLIC
CRP_b_c.version = version

EYFP_cds_c = EYFP_gene.components.create('EYFP_cds')
EYFP_cds_c.definition = EYFP_cds.persistentIdentity
EYFP_cds_c.access = SBOL_ACCESS_PUBLIC
EYFP_cds_c.version = version

EYFP_gene_constraint = EYFP_gene.sequenceConstraints.create('EYFP_gene_constraint')
EYFP_gene_constraint.subject = CRP_b_c.identity
EYFP_gene_constraint.object = EYFP_cds_c.identity
EYFP_gene_constraint.restriction = SBOL_RESTRICTION_PRECEDES

EYFP = ComponentDefinition('EYFP', BIOPAX_PROTEIN, version)
doc.addComponentDefinition(EYFP)

CRPb_circuit = ModuleDefinition('CRPb_characterization_Circuit', version)
doc.addModuleDefinition(CRPb_circuit)

cas9m_BFP_fc = CRPb_circuit.functionalComponents.create('cas9m_BFP')
cas9m_BFP_fc.definition = cas9m_BFP.identity
cas9m_BFP_fc.access = SBOL_ACCESS_PRIVATE
cas9m_BFP_fc.direction = SBOL_DIRECTION_NONE
cas9m_BFP_fc.version = version

cas9m_BFP_gene_fc = CRPb_circuit.functionalComponents.create('cas9m_BFP_gene')
cas9m_BFP_gene_fc.definition = cas9m_BFP_gene.identity
cas9m_BFP_gene_fc.access = SBOL_ACCESS_PRIVATE
cas9m_BFP_gene_fc.direction = SBOL_DIRECTION_NONE
cas9m_BFP_gene_fc.version = version

gRNA_b_fc = CRPb_circuit.functionalComponents.create('gRNA_b')
gRNA_b_fc.definition = gRNA_b.identity
gRNA_b_fc.access = SBOL_ACCESS_PRIVATE
gRNA_b_fc.direction = SBOL_DIRECTION_NONE
gRNA_b_fc.version = version

gRNA_b_gene_fc = CRPb_circuit.functionalComponents.create('gRNA_b_gene')
gRNA_b_gene_fc.definition = gRNA_b_gene.identity
gRNA_b_gene_fc.access = SBOL_ACCESS_PRIVATE
gRNA_b_gene_fc.direction = SBOL_DIRECTION_NONE
gRNA_b_gene_fc.version = version

mKate_fc = CRPb_circuit.functionalComponents.create('mKate')
mKate_fc.definition = mKate.identity
mKate_fc.access = SBOL_ACCESS_PRIVATE
mKate_fc.direction = SBOL_DIRECTION_NONE
mKate_fc.version = version

mKate_gene_fc = CRPb_circuit.functionalComponents.create('mKate_gene')
mKate_gene_fc.definition = mKate_gene.identity
mKate_gene_fc.access = SBOL_ACCESS_PRIVATE
mKate_gene_fc.direction = SBOL_DIRECTION_NONE
mKate_gene_fc.version = version

Gal4VP16_fc = CRPb_circuit.functionalComponents.create('Gal4VP16')
Gal4VP16_fc.definition = Gal4VP16.identity
Gal4VP16_fc.access = SBOL_ACCESS_PRIVATE
Gal4VP16_fc.direction = SBOL_DIRECTION_NONE
Gal4VP16_fc.version = version

Gal4VP16_gene_fc = CRPb_circuit.functionalComponents.create('Gal4VP16_gene')
Gal4VP16_gene_fc.definition = Gal4VP16_gene.identity
Gal4VP16_gene_fc.access = SBOL_ACCESS_PRIVATE
Gal4VP16_gene_fc.direction = SBOL_DIRECTION_NONE
Gal4VP16_gene_fc.version = version

EYFP_fc = CRPb_circuit.functionalComponents.create('EYFP')
EYFP_fc.definition = EYFP.identity
EYFP_fc.access = SBOL_ACCESS_PRIVATE
EYFP_fc.direction = SBOL_DIRECTION_NONE
EYFP_fc.version = version

EYFP_gene_fc = CRPb_circuit.functionalComponents.create('EYFP_gene')
EYFP_gene_fc.definition = EYFP_gene.identity
EYFP_gene_fc.access = SBOL_ACCESS_PRIVATE
EYFP_gene_fc.direction = SBOL_DIRECTION_NONE
EYFP_gene_fc.version = version

cas9m_BFP_gRNA_b_fc = CRPb_circuit.functionalComponents.create('cas9m_BFP_gRNA_b')
cas9m_BFP_gRNA_b_fc.definition = cas9m_BFP_gRNA_b.identity
cas9m_BFP_gRNA_b_fc.access = SBOL_ACCESS_PRIVATE
cas9m_BFP_gRNA_b_fc.direction = SBOL_DIRECTION_NONE
cas9m_BFP_gRNA_b_fc.version = version

mKate_production = CRPb_circuit.interactions.create('mKate_production')
mKate_production.types = [SBO_GENETIC_PRODUCTION]

mKate_participation = mKate_production.participations.create('mKate')
mKate_participation.roles = [SBO_PRODUCT]
mKate_participation.participant = mKate_fc.identity

mKate_gene_participation = mKate_production.participations.create('mKate_gene')
mKate_gene_participation.roles = [SBO_PROMOTER]
mKate_gene_participation.participant = mKate_gene_fc.identity

GAL4VP16_production = CRPb_circuit.interactions.create('Gal4VP16_production')
GAL4VP16_production.types = [SBO_GENETIC_PRODUCTION]

Gal4VP16_gene_participation = GAL4VP16_production.participations.create('Gal4VP16_gene')
Gal4VP16_gene_participation.roles = [SBO_PROMOTER]
Gal4VP16_gene_participation.participant = Gal4VP16_gene_fc.identity

Gal4VP16_participation1 = GAL4VP16_production.participations.create('Gal4VP16')
Gal4VP16_participation1.roles = [SBO_PRODUCT]
Gal4VP16_participation1.participant = Gal4VP16_fc.identity

cas9m_BFP_production = CRPb_circuit.interactions.create('cas9m_BFP_production')
cas9m_BFP_production.types = [SBO_GENETIC_PRODUCTION]

cas9m_BFP_gene_participation = \
    cas9m_BFP_production.participations.create('cas9m_BFP_gene')
cas9m_BFP_gene_participation.roles = [SBO_PROMOTER]
cas9m_BFP_gene_participation.participant = cas9m_BFP_gene_fc.identity

cas9m_BFP_participation = cas9m_BFP_production.participations.create('cas9m_BFP')
cas9m_BFP_participation.roles = [SBO_PRODUCT]
cas9m_BFP_participation.participant = cas9m_BFP_fc.identity

gRNA_b_production = CRPb_circuit.interactions.create('gRNA_b_production')
gRNA_b_production.types = [SBO_GENETIC_PRODUCTION]

gRNA_b_gene_participation = gRNA_b_production.participations.create('gRNA_b_gene')
gRNA_b_gene_participation.roles = [SBO_PROMOTER]
gRNA_b_gene_participation.participant = gRNA_b_gene_fc.identity

gRNA_b_participation = gRNA_b_production.participations.create('gRNA_b')
gRNA_b_participation.roles = [SBO_PRODUCT]
gRNA_b_participation.participant = gRNA_b_fc.identity

EYFP_Activation = CRPb_circuit.interactions.create('EYFP_Activation')
EYFP_Activation.types = [SBO_STIMULATION]
# %%
Gal4VP16_participation = EYFP_Activation.participations.create('Gal4VP16')
Gal4VP16_participation.roles = [SBO_STIMULATOR]
Gal4VP16_participation.participant = Gal4VP16_fc.identity

EYFP_gene_participation = EYFP_Activation.participations.create('EYFP_gene')
EYFP_gene_participation.roles = [SBO_PROMOTER]
EYFP_gene_participation.participant = EYFP_gene_fc.identity

mKate_deg = CRPb_circuit.interactions.create('mKate_deg')
mKate_deg.types = [SBO_DEGRADATION]

mKate_participation1 = mKate_deg.participations.create('mKate')
mKate_participation1.roles = [SBO_REACTANT]
mKate_participation1.participant = mKate_fc.identity

GAL4VP16_deg = CRPb_circuit.interactions.create('Gal4VP16_deg')
GAL4VP16_deg.types = [SBO_DEGRADATION]

Gal4VP16_participation2 = GAL4VP16_deg.participations.create('Gal4VP16')
Gal4VP16_participation2.roles = [SBO_REACTANT]
Gal4VP16_participation2.participant = Gal4VP16_fc.identity

cas9m_BFP_deg = CRPb_circuit.interactions.create('cas9m_BFP_deg')
cas9m_BFP_deg.types = [SBO_DEGRADATION]

cas9m_BFP_participation1 = cas9m_BFP_deg.participations.create('cas9m_BFP')
cas9m_BFP_participation1.roles = [SBO_REACTANT]
cas9m_BFP_participation1.participant = cas9m_BFP_fc.identity

gRNA_b_deg = CRPb_circuit.interactions.create('gRNA_b_deg')
gRNA_b_deg.types = [SBO_DEGRADATION]

gRNA_b_participation1 = gRNA_b_deg.participations.create('gRNA_b')
gRNA_b_participation1.roles = [SBO_REACTANT]
gRNA_b_participation1.participant = gRNA_b_fc.identity

EYFP_deg = CRPb_circuit.interactions.create('EYFP_deg')
EYFP_deg.types = [SBO_DEGRADATION]

EYFP_participation = EYFP_deg.participations.create('EYFP')
EYFP_participation.roles = [SBO_REACTANT]
EYFP_participation.participant = EYFP_fc.identity

cas9m_BFP_gRNA_b_deg = CRPb_circuit.interactions.create('cas9m_BFP_gRNA_b_deg')
cas9m_BFP_gRNA_b_deg.types = [SBO_DEGRADATION]

cas9m_BFP_gRNA_b_participation = \
    cas9m_BFP_gRNA_b_deg.participations.create('cas9m_BFP_gRNA_b')
cas9m_BFP_gRNA_b_participation.roles = [SBO_REACTANT]
cas9m_BFP_gRNA_b_participation.participant = cas9m_BFP_gRNA_b_fc.identity

Template_Module = CRPb_circuit.modules.create('CRISPR_Template')
Template_Module.definition = CRISPR_Template.identity

cas9m_BFP_map = Template_Module.mapsTos.create('cas9m_BFP_map')
cas9m_BFP_map.refinement = SBOL_REFINEMENT_USE_LOCAL
cas9m_BFP_map.local = cas9m_BFP_fc.identity
cas9m_BFP_map.remote = cas9_generic_fc.identity

gRNA_b_map = Template_Module.mapsTos.create('gRNA_b_map')
gRNA_b_map.refinement = SBOL_REFINEMENT_USE_LOCAL
gRNA_b_map.local = gRNA_b_fc.identity
gRNA_b_map.remote = gRNA_generic_fc.identity

cas9m_BFP_gRNA_map = Template_Module.mapsTos.create('cas9m_BFP_gRNA_map')
cas9m_BFP_gRNA_map.refinement = SBOL_REFINEMENT_USE_LOCAL
cas9m_BFP_gRNA_map.local = cas9m_BFP_gRNA_b_fc.identity
cas9m_BFP_gRNA_map.remote = cas9_gRNA_complex_fc.identity

EYFP_map = Template_Module.mapsTos.create('EYFP_map')
EYFP_map.refinement = SBOL_REFINEMENT_USE_LOCAL
EYFP_map.local = EYFP_fc.identity
EYFP_map.remote = target_fc.identity

EYFP_gene_map = Template_Module.mapsTos.create('EYFP_gene_map')
EYFP_gene_map.refinement = SBOL_REFINEMENT_USE_LOCAL
EYFP_gene_map.local = EYFP_gene_fc.identity
EYFP_gene_map.remote = target_gene_fc.identity

doc.write('crispr_example.xml')



from rdflib import URIRef

from .toplevel import TopLevel
from .constants import *
from .property import ReferencedObject


class ExperimentalData(TopLevel):
    def __init__(self, uri=URIRef('example'), version=VERSION_STRING):
        super().__init__(SBOL_EXPERIMENTAL_DATA, uri, version)


class Experiment(TopLevel):
    def __init__(self, uri=URIRef('example'), version=VERSION_STRING):
        super().__init__(SBOL_EXPERIMENT, uri, version)
        self.experimentalData = ReferencedObject(self,
                                                 SBOL_URI + "#experimentalData",
                                                 SBOL_EXPERIMENTAL_DATA,
                                                 '0', '*', [])



import datetime
import math
import unittest
import os

import sbol2

# backward compatibility
import sbol2 as sbol

import rdflib

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
TEST_LOCATION = os.path.join(MODULE_LOCATION, 'resources',
                             'crispr_example.xml')


class TestProperty(unittest.TestCase):
    def test_listProperty(self):
        plasmid = sbol.ComponentDefinition('pBB1', sbol.BIOPAX_DNA, '1.0.0')
        plasmid.roles = [sbol.SO_PLASMID, sbol.SO_CIRCULAR]
        self.assertEqual(len(plasmid.roles), 2)

    def test_noListProperty(self):
        plasmid = sbol.ComponentDefinition('pBB1', sbol.BIOPAX_DNA, '1.0.0')
        with self.assertRaises(ValueError):
            plasmid.version = ['1', '2']

    def test_addPropertyToList(self):
        plasmid = sbol.ComponentDefinition('pBB1', sbol.BIOPAX_DNA, '1.0.0')
        plasmid.roles = [sbol.SO_PLASMID]
        plasmid.addRole(sbol.SO_CIRCULAR)
        self.assertEqual(len(plasmid.roles), 2)

    def test_removePropertyFromList(self):
        plasmid = sbol.ComponentDefinition('pBB1', sbol.BIOPAX_DNA, '1.0.0')
        plasmid.roles = [sbol.SO_PLASMID, sbol.SO_CIRCULAR]
        plasmid.removeRole()
        self.assertEqual(len(plasmid.roles), 1)

    def test_unset_text_property(self):
        doc = sbol.Document()
        cd = doc.componentDefinitions.create('cd')
        cd.name = 'foo'
        self.assertEqual(cd.name, 'foo')
        cd.name = None
        self.assertEqual(cd.name, None)
        cd.name = 'foo'
        self.assertEqual(cd.name, 'foo')
        cd.name = ''
        self.assertEqual(cd.name, None)

    def test_unset_uri_property(self):
        doc = sbol.Document()
        cd = doc.componentDefinitions.create('cd')
        cd.persistentIdentity = 'foo'
        self.assertEqual('foo', cd.persistentIdentity)
        cd.persistentIdentity = None
        self.assertEqual(None, cd.persistentIdentity)
        cd.persistentIdentity = 'foo'
        self.assertEqual('foo', cd.persistentIdentity)
        cd.persistentIdentity = ''
        self.assertEqual(None, cd.persistentIdentity)

    def test_unsetListProperty(self):
        plasmid = sbol.ComponentDefinition('pBB1', sbol.BIOPAX_DNA, '1.0.0')
        plasmid.roles = [sbol.SO_PLASMID, sbol.SO_CIRCULAR]
        plasmid.roles = []
        self.assertEqual(len(plasmid.roles), 0)

    def test_lenOwnedObject(self):
        d = sbol.Document()
        d.read(TEST_LOCATION)
        self.assertEqual(25, len(d.componentDefinitions))
        self.assertEqual(2, len(d.moduleDefinitions))
        self.assertEqual(4, len(d.sequences))

    def test_getitem(self):
        sbol.setHomespace('http://sbols.org/CRISPR_Example/')
        d = sbol.Document()
        d.read(TEST_LOCATION)
        s1 = d.sequences['CRa_U6_seq']
        expected = 'http://sbols.org/CRISPR_Example/CRa_U6_seq/1.0.0'
        self.assertEqual(expected, str(s1))

    def test_readProperties(self):
        d = sbol.Document()
        d.read(TEST_LOCATION)
        cd2 = d.componentDefinitions['http://sbols.org/CRISPR_Example/EYFP_gene/1.0.0']
        self.assertEqual(2, len(cd2.components))
        self.assertEqual(1, len(cd2.roles))
        md = d.moduleDefinitions['http://sbols.org/CRISPR_Example/CRISPR_Template/1.0.0']
        self.assertEqual(5, len(md.functionalComponents))
        self.assertEqual(3, len(md.interactions))
        self.assertEqual(0, len(md.roles))

    def test_text_property_constructor(self):
        # Test None as parent object
        with self.assertRaises(AttributeError):
            sbol.TextProperty(None, sbol.SBOL_NAME, '0', '*', [], 'foo')
        # Test string as parent object
        with self.assertRaises(AttributeError):
            sbol.TextProperty('foo', sbol.SBOL_NAME, '0', '*', [], 'foo')
        # Test with object whose properties attribute is not a dict
        with self.assertRaises(TypeError):
            md = sbol.ModuleDefinition()
            md.properties = []
            sbol.TextProperty(md, sbol.SBOL_NAME, '0', '*', [], 'foo')

    def test_literal_property_properties(self):
        md = sbol.ModuleDefinition()
        self.assertNotIn(rdflib.URIRef(sbol.UNDEFINED), md.properties)
        sbol.property.TextProperty(md, sbol.UNDEFINED, '0', '*', [], 'foo')
        # Creating the property should also create the entry in the
        # parent properties dict
        self.assertIn(rdflib.URIRef(sbol.UNDEFINED), md.properties)

    def test_text_property_setting_single(self):
        md = sbol.ModuleDefinition()
        testing_uri = rdflib.URIRef(sbol.SBOL_URI + "#Testing")
        tp = sbol.TextProperty(md, testing_uri, '0', '1', [])
        # Test setting to string
        expected = 'foo'
        tp.value = expected
        self.assertEqual(tp.value, expected)
        # Test setting to None
        tp.value = None
        self.assertIsNone(tp.value)
        # Test integer
        with self.assertRaises(TypeError):
            tp.value = 3
        # Test setting to list
        with self.assertRaises(TypeError):
            tp.value = ['foo', 'bar']

    def test_text_property_setting_list(self):
        md = sbol.ModuleDefinition()
        testing_uri = rdflib.URIRef(sbol.SBOL_URI + "#Testing")
        tp = sbol.TextProperty(md, testing_uri, '0', '*', [])
        # Test setting to string
        expected = 'foo'
        tp.value = expected
        self.assertEqual(tp.value, [expected])
        # Test setting to None
        with self.assertRaises(TypeError):
            tp.value = None
        # Test setting to list
        expected = ['foo', 'bar']
        tp.value = expected
        self.assertEqual(tp.value, expected)
        # Test setting to list of integers
        with self.assertRaises(TypeError):
            tp.value = [1, 2, 3]
        # Test setting to empty list
        expected = []
        tp.value = expected
        self.assertEqual(tp.value, [])

    def test_owned_object_singleton(self):
        cd = sbol.ComponentDefinition('cd')
        annotation_uri = rdflib.URIRef('http://examples.org#annotation_property')
        cd.annotation = sbol.property.OwnedObject(cd, annotation_uri, sbol.Identified,
                                                  '0', '1', None)
        self.assertIsNone(cd.annotation)
        cd.annotation = sbol.Identified('foo')
        self.assertEqual(type(cd.annotation), sbol.Identified)

        # Test unsetting
        cd.annotation = None
        self.assertEqual(cd.annotation, None)
        cd.annotation = None
        self.assertEqual(cd.annotation, None)

    def test_owned_object_singleton_ints(self):
        # same as test_owned_object_singleton but specify OwnedObject
        # bounds as ints instead of strings
        cd = sbol.ComponentDefinition('cd')
        annotation_uri = rdflib.URIRef('http://examples.org#annotation_property')
        cd.annotation = sbol.property.OwnedObject(cd, annotation_uri, sbol.Identified,
                                                  0, 1, None)
        self.assertIsNone(cd.annotation)
        cd.annotation = sbol.Identified('foo')
        self.assertEqual(type(cd.annotation), sbol.Identified)

        # Test unsetting
        cd.annotation = None
        self.assertEqual(cd.annotation, None)
        cd.annotation = None
        self.assertEqual(cd.annotation, None)

    def test_owned_object_multiple(self):
        cd = sbol.ComponentDefinition('cd')
        annotation_uri = rdflib.URIRef('http://examples.org#annotation_property')
        cd.annotations = sbol.property.OwnedObject(cd, annotation_uri, sbol.Identified,
                                                   '0', '*', None)
        self.assertEqual(type(cd.annotations), sbol.property.OwnedObject)
        cd.annotations.add(sbol.Identified(uri='foo_0'))
        cd.annotations.add(sbol.Identified(uri='foo_1'))
        self.assertEqual(type(cd.annotations), sbol.property.OwnedObject)

    def test_owned_object_find(self):
        doc = sbol.Document()
        md = doc.moduleDefinitions.create('foo')
        # find() underlies __contains__ so test `in`
        self.assertIn('foo', doc.moduleDefinitions)
        self.assertNotIn('bar', doc.moduleDefinitions)
        # find something that is in the collection
        md2 = doc.moduleDefinitions.find('foo')
        self.assertEqual(md, md2)
        # Try to find something that is not in the collection,
        # which should return False as pysbol did
        self.assertFalse(doc.moduleDefinitions.find('bar'))

    def test_referenced_object(self):
        # Test referenced object property is initialized to correct types
        cd0 = sbol.ComponentDefinition('cd0')
        self.assertEqual(type(cd0.sequences), list)

        c = cd0.components.create('c')
        self.assertEqual(c.definition, None)

        # Test assignment
        cd1 = sbol.ComponentDefinition('cd1')
        c.definition = cd1.identity
        self.assertEqual(c.definition, cd1.identity)

        seq0a = sbol.Sequence('seq0a')
        seq0b = sbol.Sequence('seq0b')
        cd0.sequences = [seq0a.identity, seq0b.identity]
        self.assertEqual(cd0.sequences, [seq0a.identity, seq0b.identity])

        c.definition = cd1
        self.assertEqual(c.definition, cd1.identity)

        # Test conversion to URIRef
        c.definition = str(cd1.identity)
        self.assertEqual(type(c.definition), str)

        cd0.sequences = [str(seq0a.identity), str(seq0b.identity)]
        self.assertEqual([type(s) for s in cd0.sequences],
                         [str, str])

        # Test unset
        c.definition = None
        self.assertEqual(c.definition, None)

        c.definition = cd1.identity
        c.definition = ''
        self.assertEqual(c.definition, None)

        cd0.sequences = []
        self.assertEqual(cd0.sequences, [])

        cd0.sequences = [seq0a.identity, seq0b.identity]
        cd0.sequences = None
        self.assertEqual(cd0.sequences, [])

        with self.assertRaises(TypeError):
            cd0.sequences = [seq0a.identity, seq0b.identity]
            cd0.sequences = [None, None]
            self.assertEqual(cd0.sequences, [])

    def test_int_property(self):
        r = sbol.location.Range('r')
        r.start = 42
        self.assertEqual(type(r.start), int)
        self.assertEqual(r.start, 42)
        with self.assertRaises(ValueError):
            r.start = 'forty-two'

    def test_uri_property_list(self):
        cd = sbol.ComponentDefinition('cd')
        foo_str = 'foo'
        bar_str = 'bar'
        foo_uri = rdflib.URIRef(foo_str)
        bar_uri = rdflib.URIRef(bar_str)
        cd.roles = foo_str
        self.assertEqual(cd.roles, [foo_str])
        # This used to append. Eek!
        cd.roles = bar_str
        self.assertEqual(cd.roles, [bar_str])
        # This used to append. Eek!
        cd.roles = [foo_str, bar_str]
        self.assertEqual(cd.roles, [foo_str, bar_str])
        with self.assertRaises(TypeError):
            cd.roles = 34

    def test_uri_property_append(self):
        cd = sbol.ComponentDefinition('cd')
        foo = 'foo'
        bar = 'bar'
        cd.roles = foo
        self.assertEqual(cd.roles, [foo])
        # append has no effect. This is because the list that is
        # returned by `cd.roles` is not connected with the internal
        # data representation.
        cd.roles.append(bar)
        self.assertEqual(cd.roles, [foo])
        # The `+=` operator is a good way for users to append new
        # items to the list
        cd.roles += [bar]
        self.assertEqual(cd.roles, [foo, bar])

    def test_owned_object_remove(self):
        md = sbol.ModuleDefinition('md')
        m1 = md.modules.create('m1')
        m2 = md.modules.create('m2')
        self.assertEqual(list(md.modules), [m1, m2])
        obj = md.modules.remove(m2.identity)
        self.assertEqual(list(md.modules), [m1])
        self.assertEqual(obj.identity, m2.identity)
        self.assertIsNone(obj.doc)
        self.assertIsNone(m2.doc)

    def test_validation_rules(self):
        md = sbol2.ModuleDefinition()
        with self.assertRaises(TypeError):
            # No validation rules, so `'AND'` is interpreted
            # as validation rules.
            tp = sbol2.TextProperty(md, 'http://example.com#logic', '0', '1',
                                    'AND', 'initial_value')
        # Use an empty list to specify no validation rules
        tp = sbol2.TextProperty(md, 'http://example.com#logic', '0', '1',
                                [], 'AND')
        # Should add a test for callable validation rules

    def test_upper_bound_math_inf(self):
        # test that math.inf can be used as an upper bound
        cd = sbol.ComponentDefinition('cd')
        annotation_uri = rdflib.URIRef('http://examples.org#annotation_property')
        cd.annotation = sbol.property.OwnedObject(cd, annotation_uri, sbol.Identified,
                                                  0, math.inf, None)
        # By getting here, the test has passed
        self.assertEqual(0, len(cd.annotation))

    def test_lower_bound_math_inf(self):
        # test that math.inf can not be used as a lower bound
        cd = sbol.ComponentDefinition('cd')
        annotation_uri = rdflib.URIRef('http://examples.org#annotation_property')
        with self.assertRaises(ValueError):
            cd.annotation = sbol.property.OwnedObject(cd, annotation_uri,
                                                      sbol.Identified,
                                                      math.inf, 1, None)


class TestIntProperty(unittest.TestCase):

    def test_four_arg_constructor(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testInt'
        prop = sbol2.IntProperty(cd, type_uri, '0', '1')
        self.assertEqual(None, prop.value)
        self.assertEqual([], prop._validation_rules)

    def test_five_arg_constructor_validators(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testInt'
        validation_rules = [print]
        prop = sbol2.IntProperty(cd, type_uri, '0', '*', validation_rules)
        self.assertEqual([], prop.value)
        self.assertEqual(validation_rules, prop._validation_rules)

    def test_five_arg_constructor_one_value(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testInt'
        initial_value = 2
        prop = sbol2.IntProperty(cd, type_uri, '0', '1', initial_value)
        self.assertEqual(initial_value, prop.value)
        self.assertEqual([], prop._validation_rules)

    def test_five_arg_constructor_values(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testInt'
        initial_value = [1, 2, 3]
        prop = sbol2.IntProperty(cd, type_uri, '0', '*', initial_value)
        self.assertEqual(initial_value, prop.value)
        self.assertEqual([], prop._validation_rules)

    def test_six_arg_constructor(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testInt'
        initial_value = 32
        prop = sbol2.IntProperty(cd, type_uri, '0', '1', [], initial_value)
        self.assertEqual(initial_value, prop.value)
        self.assertEqual([], prop._validation_rules)

    def test_six_arg_constructor_none(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testInt'
        initial_value = 32
        prop = sbol2.IntProperty(cd, type_uri, '0', '1', None, initial_value)
        self.assertEqual(initial_value, prop.value)
        self.assertEqual([], prop._validation_rules)

    def test_values(self):
        # Cut class uses IntProperty
        # Make sure we can set a IntProperty via a variety of types
        # that can be coerced to float
        cut = sbol2.Cut('m1')
        self.assertEqual(0, cut.at)
        # set with float
        v = 2.54
        cut.at = v
        self.assertEqual(int(v), cut.at)
        # set with string
        v = '32'
        cut.at = v
        self.assertEqual(int(v), cut.at)
        # set with string float
        v = '1.37'
        with self.assertRaises(ValueError):
            cut.at = v
        # set with int
        v = 15
        cut.at = v
        self.assertEqual(int(v), cut.at)

    def test_init_store(self):
        # Ensure that property constructors initialize the parent
        # object's value store
        obj = sbol2.SBOLObject()
        type_uri = 'http://example.com#thing'
        obj.thing = sbol2.IntProperty(obj, type_uri, '0', '*')
        self.assertIn(type_uri, obj.properties)
        self.assertEqual([], obj.properties[type_uri])
        self.assertEqual([], obj.thing)


class TestTextProperty(unittest.TestCase):

    def test_four_arg_constructor(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testText'
        prop = sbol2.TextProperty(cd, type_uri, '0', '1')
        self.assertEqual(None, prop.value)
        self.assertEqual([], prop._validation_rules)

    def test_five_arg_constructor_validators(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testText'
        validation_rules = [print]
        prop = sbol2.TextProperty(cd, type_uri, '0', '*', validation_rules)
        self.assertEqual([], prop.value)
        self.assertEqual(validation_rules, prop._validation_rules)

    def test_five_arg_constructor_one_value(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testText'
        initial_value = 'omega'
        prop = sbol2.TextProperty(cd, type_uri, '0', '1', initial_value)
        self.assertEqual(initial_value, prop.value)
        self.assertEqual([], prop._validation_rules)

    def test_five_arg_constructor_values(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testText'
        initial_value = ['alpha', 'bravo', 'charlie']
        prop = sbol2.TextProperty(cd, type_uri, '0', '*', initial_value)
        self.assertEqual(initial_value, prop.value)
        self.assertEqual([], prop._validation_rules)

    def test_six_arg_constructor(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testText'
        initial_value = 'omega'
        prop = sbol2.TextProperty(cd, type_uri, '0', '1', [], initial_value)
        self.assertEqual(initial_value, prop.value)
        self.assertEqual([], prop._validation_rules)

    def test_six_arg_constructor_none(self):
        cd = sbol2.ComponentDefinition()
        type_uri = 'http://example.com/test#testText'
        initial_value = 'omega'
        prop = sbol2.TextProperty(cd, type_uri, '0', '1', None, initial_value)
        self.assertEqual(initial_value, prop.value)
        self.assertEqual([], prop._validation_rules)

    def test_init_store(self):
        # Ensure that property constructors initialize the parent
        # object's value store
        obj = sbol2.SBOLObject()
        type_uri = 'http://example.com#thing'
        obj.thing = sbol2.TextProperty(obj, type_uri, '0', '*')
        self.assertIn(type_uri, obj.properties)
        self.assertEqual([], obj.properties[type_uri])
        self.assertEqual([], obj.thing)


class TestFloatProperty(unittest.TestCase):

    def test_values(self):
        # Measurement class uses FloatProperty
        # Make sure we can set a FloatProperty via a variety of types
        # that can be coerced to float
        m = sbol2.Measurement('m1')
        self.assertEqual(0.0, m.value)
        # set with float
        v = 2.54
        m.value = v
        self.assertEqual(v, m.value)
        # set with string
        v = '1.37'
        m.value = v
        self.assertEqual(float(v), m.value)
        # set with int
        v = 15
        m.value = v
        self.assertEqual(float(v), m.value)

    def test_init_store(self):
        # Ensure that property constructors initialize the parent
        # object's value store
        obj = sbol2.SBOLObject()
        type_uri = 'http://example.com#thing'
        obj.thing = sbol2.FloatProperty(obj, type_uri, '0', '*')
        self.assertIn(type_uri, obj.properties)
        self.assertEqual([], obj.properties[type_uri])
        self.assertEqual([], obj.thing)


class TestDateTimeProperty(unittest.TestCase):

    def test_values(self):
        # Activity class uses DateTimeProperty
        # Make sure we can set a DateTimeProperty via a variety of formats
        # that can be parsed by dateutil
        activity = sbol2.Activity('a1')
        self.assertIsNone(activity.startedAtTime)
        # set with datetime
        v = datetime.datetime.now()
        activity.startedAtTime = v
        self.assertEqual(v, activity.startedAtTime)
        # set with string
        dt = datetime.datetime.now()
        v = str(dt)
        activity.startedAtTime = v
        self.assertEqual(dt, activity.startedAtTime)
        # set with string in ISO format
        dt = datetime.datetime.now()
        v = dt.isoformat()
        activity.startedAtTime = v
        self.assertEqual(dt, activity.startedAtTime)

    def test_init_store(self):
        # Ensure that property constructors initialize the parent
        # object's value store
        obj = sbol2.SBOLObject()
        type_uri = 'http://example.com#thing'
        obj.thing = sbol2.DateTimeProperty(obj, type_uri, '0', '*')
        self.assertIn(type_uri, obj.properties)
        self.assertEqual([], obj.properties[type_uri])
        self.assertEqual([], obj.thing)


class TestReferencedObject(unittest.TestCase):

    def test_init_store(self):
        # Ensure that property constructors initialize the parent
        # object's value store
        obj = sbol2.SBOLObject()
        type_uri = 'http://example.com#thing'
        ref_uri = 'http://example.com#other_thing'
        obj.thing = sbol2.ReferencedObject(obj, type_uri, ref_uri, '0', '*', [])
        self.assertIn(type_uri, obj.properties)
        self.assertEqual([], obj.properties[type_uri])
        self.assertEqual([], obj.thing)


class TestURIProperty(unittest.TestCase):

    def test_init_store(self):
        # Ensure that property constructors initialize the parent
        # object's value store
        obj = sbol2.SBOLObject()
        type_uri = 'http://example.com#thing'
        obj.thing = sbol2.URIProperty(obj, type_uri, '0', '*', [])
        self.assertIn(type_uri, obj.properties)
        self.assertEqual([], obj.properties[type_uri])
        self.assertEqual([], obj.thing)


if __name__ == '__main__':
    unittest.main()



import io
import locale
import os
import tempfile
import unittest
import unittest.mock

import rdflib

import sbol2
import sbol2 as sbol

MODULE_LOCATION = os.path.dirname(os.path.abspath(__file__))
TEST_LOCATION = os.path.join(MODULE_LOCATION, 'resources', 'crispr_example.xml')
CRISPR_LOCATION = TEST_LOCATION

# For testing reading of annotations
ANNO_LOCATION = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL2',
                             'AnnotationOutput.xml')


class TestDocument(unittest.TestCase):

    def setUp(self):
        # One of the unit tests changes the validator URL
        # configuration, so we capture it here and reset it in
        # tearDown(). It would be nice to have a method to reset to
        # factory configuration.
        self.validator_url = sbol2.Config.getOption(sbol2.ConfigOptions.VALIDATOR_URL)

    def tearDown(self):
        sbol2.Config.setOption(sbol2.ConfigOptions.VALIDATOR_URL,
                               self.validator_url)

    def test_empty_len0(self):
        doc = sbol.Document()
        # print(doc)
        self.assertEqual(0, len(doc), "Length of document should be 0")

    def test_addGetTopLevel_uri(self):
        doc = sbol.Document()
        # Tutorial doesn't drop final forward slash, but this isn't right.
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption(sbol2.ConfigOptions.SBOL_COMPLIANT_URIS, True)
        sbol.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        cas9 = sbol.ComponentDefinition('Cas9', sbol.BIOPAX_PROTEIN)
        doc.addModuleDefinition(crispr_template)
        doc.addComponentDefinition(cas9)

        # Note: tutorial has 1.0.0 instead of 1 but this doesn't work
        crispr_template_uri = 'http://sbols.org/CRISPR_Example/CRISPR_Template/1'
        crispr_template_2 = doc.getModuleDefinition(crispr_template_uri)
        cas9_2 = doc.getComponentDefinition('http://sbols.org/CRISPR_Example/Cas9/1')
        self.assertEqual(crispr_template, crispr_template_2)
        self.assertEqual(cas9, cas9_2)

    def test_addGetTopLevel_displayId(self):
        doc = sbol.Document()
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption(sbol2.ConfigOptions.SBOL_COMPLIANT_URIS, True)
        sbol.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        cas9 = sbol.ComponentDefinition('Cas9', sbol.BIOPAX_PROTEIN)
        doc.addModuleDefinition(crispr_template)
        doc.addComponentDefinition(cas9)

        crispr_template_2 = doc.moduleDefinitions['CRISPR_Template']
        cas9_2 = doc.componentDefinitions['Cas9']
        self.assertEqual(crispr_template, crispr_template_2)
        self.assertEqual(cas9, cas9_2)

    def test_addGetTopLevel_indexing(self):
        doc = sbol.Document()
        # Tutorial doesn't drop final forward slash, but this isn't right.
        sbol.setHomespace('http://sbols.org/CRISPR_Example')
        sbol.Config.setOption(sbol2.ConfigOptions.SBOL_COMPLIANT_URIS, True)
        sbol.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, False)
        crispr_template = sbol.ModuleDefinition('CRISPR_Template')
        cas9 = sbol.ComponentDefinition('Cas9', sbol.BIOPAX_PROTEIN)
        doc.addModuleDefinition(crispr_template)
        doc.addComponentDefinition(cas9)

        crispr_template_2 = doc.moduleDefinitions[0]
        cas9_2 = doc.componentDefinitions[0]
        self.assertEqual(crispr_template, crispr_template_2)
        self.assertEqual(cas9, cas9_2)

    def test_iteration(self):
        doc = sbol.Document()
        doc.read(TEST_LOCATION)
        i = 0
        for obj in doc:
            i += 1
            # print(obj)
        self.assertEqual(len(doc), 31)
        # print(doc)

    def test_identity(self):
        # The sbol:identity relation should not be written out when
        # serializing SBOL.
        doc = sbol.Document()
        doc.read(TEST_LOCATION)
        result = doc.writeString()
        self.assertNotIn('sbol:identity', result)

    def test_utf8_append(self):
        utf8_path = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL2',
                                 'pICSL50014.xml')
        doc = sbol.Document()
        doc.append(utf8_path)

    def test_utf8_append_no_locale(self):
        # Test loading a utf-8 SBOL file without LANG set. This was a
        # bug at one time, and only shows itself when LANG is unset.
        # Here we simulate that by temporarily setting the locale to
        # the generic 'C' locale.
        utf8_path = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL2',
                                 'pICSL50014.xml')
        loc = locale.getlocale()
        try:
            locale.setlocale(locale.LC_ALL, 'C')
            doc = sbol.Document()
            doc.append(utf8_path)
        finally:
            locale.setlocale(locale.LC_ALL, loc)

    def test_utf8_read(self):
        utf8_path = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL2',
                                 'pICSL50014.xml')
        doc = sbol.Document()
        doc.read(utf8_path)

    def test_utf8_read_no_locale(self):
        # Test loading a utf-8 SBOL file without LANG set. This was a
        # bug at one time, and only shows itself when LANG is unset.
        # Here we simulate that by temporarily setting the locale to
        # the generic 'C' locale.
        utf8_path = os.path.join(MODULE_LOCATION, 'SBOLTestSuite', 'SBOL2',
                                 'pICSL50014.xml')
        loc = locale.getlocale()
        try:
            locale.setlocale(locale.LC_ALL, 'C')
            doc = sbol.Document()
            doc.read(utf8_path)
        finally:
            locale.setlocale(locale.LC_ALL, loc)

    def test_constructor(self):
        doc = sbol.Document(TEST_LOCATION)
        self.assertEqual(len(doc), 31)
        self.assertEqual(len(doc.componentDefinitions), 25)

    def test_find(self):
        doc = sbol.Document(TEST_LOCATION)
        found = doc.find('http://sbols.org/CRISPR_Example/CRISPR_Template')
        # At one point Document.find was returning -1 because it was calling str.find
        self.assertNotEqual(found, -1)
        self.assertIsNone(found)
        found = doc.find('http://sbols.org/CRISPR_Example/CRISPR_Template/1.0.0')
        self.assertNotEqual(found, -1)
        self.assertIsNotNone(found)

    def test_lookup(self):
        # Test simple key lookup via the __getitem__() method
        doc = sbol.Document()
        md = doc.moduleDefinitions.create('foo')
        md2 = doc.moduleDefinitions[md.displayId]
        self.assertEqual(md.identity, md2.identity)

    def test_crispr_lookup(self):
        doc = sbol.Document(TEST_LOCATION)
        uri = 'http://sbols.org/CRISPR_Example/CRISPR_Template/1.0.0'
        display_id = 'CRISPR_Template'
        md = doc.moduleDefinitions[uri]
        self.assertIsNotNone(md)
        self.assertEqual(md.identity, uri)
        self.assertEqual(md.displayId, display_id)
        # Test lookup by displayId. This was broken when loading from a file.
        md = doc.moduleDefinitions[display_id]
        self.assertIsNotNone(md)
        self.assertEqual(md.identity, uri)
        self.assertEqual(md.displayId, display_id)

    def test_find_property_value(self):
        # find_property_value wasn't comparing against the passed
        # `value` parameter, so it was returning all identities in the
        # document.
        sbol.setHomespace('http://examples.org')
        sbol.Config.setOption(sbol2.ConfigOptions.SBOL_COMPLIANT_URIS, True)
        sbol.Config.setOption(sbol2.ConfigOptions.SBOL_TYPED_URIS, True)
        doc = sbol.Document()
        cd = doc.moduleDefinitions.create('foo')
        cd.roles = sbol.SO_PROMOTER
        test_uri = 'http://examples.org/does/not/exist/1'
        matches = doc.find_property_value(sbol.SBOL_IDENTITY, test_uri)
        self.assertEqual(len(matches), 0)
        matches = doc.find_property_value(sbol.SBOL_ROLES, sbol.SO_PROMOTER)
        self.assertEqual(len(matches), 1)
        self.assertEqual(matches[0], cd)

    def test_add_namespace(self):
        doc = sbol.Document()
        cd = doc.componentDefinitions.create('cd')
        cd.foo = sbol.property.TextProperty(cd, 'http://examples.org#foo',
                                            '0', '1', None, 'bar')
        doc.readString(doc.writeString())
        namespaces = [n[1] for n in doc.graph.namespace_manager.namespaces()]
        self.assertFalse('http://examples.org#' in namespaces)
        doc.addNamespace('http://examples.org#', 'examples')
        doc.readString(doc.writeString())
        namespaces = [n for n in doc.graph.namespace_manager.namespaces()]
        self.assertIn(('examples', rdflib.URIRef('http://examples.org#')),
                      namespaces)

    def test_namespace_fail(self):
        doc = sbol.Document()
        with self.assertRaises(ValueError):
            doc.addNamespace('http://examples.org', 'foo')

    def test_doc(self):
        doc = sbol.Document()
        # This was actually broken, so don't laugh at it
        self.assertEqual(doc.doc, doc)
        md = doc.moduleDefinitions.create('foo')
        self.assertEqual(doc, md.doc)
        m = md.modules.create('bar')
        self.assertEqual(doc, m.doc)

    def test_compare(self):
        doc = sbol.Document()
        doc2 = sbol.Document()
        self.assertTrue(doc.compare(doc2))
        doc.addNamespace('http://example.org#', 'bar')
        self.assertFalse(doc.compare(doc2))

    def test_get_top_level(self):
        doc = sbol.Document()
        self.assertTrue(hasattr(doc, 'getTopLevel'))
        # Expect a not found error in an empty document
        try:
            doc.getTopLevel('foo')
        except sbol.SBOLError as err:
            self.assertEqual(err.error_code(),
                             sbol.SBOLErrorCode.SBOL_ERROR_NOT_FOUND)
        else:
            self.fail('Expected SBOLError')
        cd = doc.componentDefinitions.create('foo')
        cd2 = doc.getTopLevel(cd.identity)
        self.assertEqual(cd, cd2)
        # Test access via string. Above was via URIRef.
        cd2 = doc.getTopLevel(str(cd.identity))
        self.assertEqual(cd, cd2)

    def test_addImplementation(self):
        doc = sbol.Document()
        with self.assertRaises(Exception):
            # This raises an attribute exception right now, and that
            # should change in the future.
            doc.addImplementation('foo')
        impl = sbol.Implementation(uri='foo')
        doc.addImplementation(impl)
        self.assertEqual(doc.implementations[0], impl)
        impl2 = sbol.Implementation(uri='bar')
        impl3 = sbol.Implementation(uri='baz')
        doc.addImplementation([impl2, impl3])
        self.assertEqual(doc.implementations[1], impl2)
        self.assertEqual(doc.implementations[2], impl3)

    def test_addImplementationGenerator(self):
        # We should be able to add implementations from a generator.
        # We should not be restricted to adding only a list. Any
        # Iterable should be fine.
        def impl_gen(count):
            for i in range(count):
                uri = 'impl{}'.format(i)
                yield sbol.Implementation(uri=uri)
        homespace = 'http://sbols.org/sbol_test'
        sbol.setHomespace(homespace)
        doc = sbol.Document()
        doc.addImplementation(impl_gen(3))
        uri_template = '{}/impl{}/1'
        self.assertEqual(doc.implementations[0].identity,
                         uri_template.format(homespace, '0'))
        self.assertEqual(doc.implementations[1].identity,
                         uri_template.format(homespace, '1'))
        self.assertEqual(doc.implementations[2].identity,
                         uri_template.format(homespace, '2'))

    def test_getImplementation(self):
        homespace = 'http://sbols.org/sbol_test'
        sbol.setHomespace(homespace)
        doc = sbol.Document()
        impl = sbol.Implementation(uri='foo')
        doc.addImplementation(impl)
        impl2 = sbol.Implementation(uri='bar')
        doc.addImplementation(impl2)
        # Get by index
        self.assertEqual(doc.getImplementation(0), impl)
        self.assertEqual(doc.getImplementation(1), impl2)
        # Get by short URI
        self.assertEqual(doc.getImplementation('foo'), impl)
        self.assertEqual(doc.getImplementation('bar'), impl2)
        # Get by full URI
        uri_template = '{}/Implementation/{}/1'
        uri = rdflib.URIRef(uri_template.format(homespace, 'foo'))
        self.assertEqual(doc.getImplementation(uri), impl)
        uri = rdflib.URIRef(uri_template.format(homespace, 'bar'))
        self.assertEqual(doc.getImplementation(uri), impl2)

    def test_bool(self):
        doc = sbol.Document()
        self.assertTrue(doc)

    def test_from_user_to_user(self):
        # Test proper conversion of user-facing types from RDFlib types
        # through the Property interfaces after serializing and de-serializing
        doc = sbol.Document()
        cd = doc.componentDefinitions.create('cd')
        cd.int_property = sbol.IntProperty(cd, 'http://examples.org', '0', '1', None, 42)
        self.assertEqual(42, cd.int_property)
        doc2 = sbol.Document()
        doc2.readString(doc.writeString())
        cd2 = doc2.componentDefinitions['cd']
        cd2.int_property = sbol.IntProperty(cd, 'http://examples.org', '0', '1', None)
        self.assertEqual(42, cd2.int_property)

    def test_range(self):
        # Test proper serializing and de-serializing of range properties
        doc = sbol.Document()
        cd = doc.componentDefinitions.create('cd')
        sa = cd.sequenceAnnotations.create('sa')
        r = sa.locations.createRange('r')
        r.start = 42
        doc2 = sbol.Document()
        doc2.readString(doc.writeString())
        r = doc2.componentDefinitions['cd'].sequenceAnnotations['sa'].locations['r']
        self.assertEqual(r.start, 42)

    def test_clone_document(self):
        doc = sbol.Document()
        doc2 = doc.copy()
        self.assertTrue(doc.compare(doc2))

    def test_validate(self):
        doc = sbol2.Document()
        # Add a module definition
        doc.moduleDefinitions.create('md')
        result = doc.validate()
        expected = 'Valid.'
        self.assertEqual(result, expected)

    def test_validate_bad_url(self):
        sbol2.Config.setOption(sbol2.ConfigOptions.VALIDATOR_URL,
                               self.validator_url + 'foo/bar')
        doc = sbol2.Document()
        with self.assertRaises(sbol2.SBOLError):
            doc.validate()

    def test_read_annotations(self):
        # Test reading a file with annotations and make sure they end
        # up where we expect them
        doc = sbol2.Document(filename=ANNO_LOCATION)
        # There is 1 component definition
        self.assertEqual(len(doc.componentDefinitions), 1)
        cd = doc.componentDefinitions[0]
        info_uri = rdflib.URIRef('http://partsregistry.org/information')
        sigma_uri = rdflib.URIRef('http://partsregistry.org/sigmafactor')
        regulation_uri = rdflib.URIRef('http://partsregistry.org/regulation')
        self.assertNotIn(info_uri, cd.properties)
        self.assertIn(info_uri, cd.owned_objects)
        # TODO What is the real API for accessing the extension object?
        info = cd.owned_objects[info_uri][0]
        self.assertIsNotNone(info)
        self.assertEqual(info.getPropertyValue(sigma_uri),
                         '//rnap/prokaryote/ecoli/sigma70')
        self.assertEqual(info.getPropertyValue(regulation_uri),
                         '//regulation/constitutive')

    def test_recursive_add(self):
        # Make sure that when an object gets added to a document
        # all of its child objects also get added.
        cd = sbol2.ComponentDefinition('cd')
        comp = sbol2.Component('cd_c')
        cd.components.add(comp)
        # Use of cd.sequence is dubious because the sequence attribute
        # isn't really there in SBOL 2.3. But it's the test case that
        # found the bug with recursive addition of objects, so we use it.
        seq = sbol2.Sequence('cd_seq')
        cd.sequence = seq
        doc = sbol2.Document()
        doc.addComponentDefinition(cd)
        # The cd and sequence should be in the document
        # The component is not top level, so doesn't get added
        self.assertEqual(2, len(doc))

    def test_add_attachment(self):
        doc = sbol2.Document()
        test_attach = sbol2.Attachment("TEST")
        doc.addAttachment(test_attach)
        self.assertEqual(1, len(doc.attachments))
        self.assertTrue(test_attach.compare(doc.attachments[0]))

    def test_get_collection(self):
        doc = sbol2.Document()
        self.assertTrue(hasattr(doc, 'getCollection'))
        c1 = doc.collections.create('c1')
        self.assertEqual(c1, doc.getCollection('c1'))
        self.assertEqual(c1, doc.getCollection(c1.identity))

    def test_read_string_clear(self):
        # Test that Document.readString() clears the document
        doc = sbol2.Document()
        cd = doc.componentDefinitions.create('test_cd')
        _ = cd.components.create('test_c')
        self.assertEqual(1, len(cd.components))
        doc.readString(doc.writeString())
        self.assertEqual(1, len(cd.components))
        cd = doc.componentDefinitions[cd.identity]
        self.assertEqual(1, len(cd.components))


class NonTopLevelExtension(sbol2.Identified):

    RDF_TYPE = 'http://example.org/test#NonTopLevelExtension'

    def __init__(self, uri='example'):
        super().__init__(uri=uri,
                         type_uri=NonTopLevelExtension.RDF_TYPE)


class TopLevelExtension(sbol2.TopLevel):

    RDF_TYPE = 'http://example.org/test#TopLevelExtension'

    def __init__(self, uri='example'):
        super().__init__(uri=uri,
                         type_uri=TopLevelExtension.RDF_TYPE)
        self.child = sbol2.OwnedObject(self, 'http://example.org/test#child',
                                       NonTopLevelExtension, '0', '1', [])


class ComponentDefinitionOverride(sbol2.ComponentDefinition):

    def __init__(self, uri='example'):
        super().__init__(uri=uri)


class TestDocumentExtensionObjects(unittest.TestCase):

    def test_register_extension_class(self):

        # Define extension object
        doc = sbol2.Document()
        cd = ComponentDefinitionOverride('cd')
        doc.add(cd)

        # Round-trip the extension data
        doc2 = sbol2.Document()
        doc2.readString(doc.writeString())
        cd = doc2.getExtensionObject(cd.identity)

        # Note the extension object's type is not preserved!
        self.assertIs(type(cd), sbol2.ComponentDefinition)
        self.assertIsNot(type(cd), ComponentDefinitionOverride)

        # Now register the class and attempt to round-trip again
        doc2 = sbol2.Document()
        sbol2.Config.register_extension_class(ComponentDefinitionOverride,
                                              sbol2.SBOL_COMPONENT_DEFINITION)
        doc2.readString(doc.writeString())
        cd = doc2.getExtensionObject(cd.identity)

        # This time the extension object's type is preserved!
        self.assertIs(type(cd), ComponentDefinitionOverride)
        self.assertIsNot(type(cd), sbol2.ComponentDefinition)

        # The object is stored in Document.componentDefinitions
        self.assertEqual(len(doc2.componentDefinitions), 1)

        # Restore
        sbol2.Config.register_extension_class(sbol2.ComponentDefinition,
                                              sbol2.SBOL_COMPONENT_DEFINITION)

    def test_get_extension_object(self):
        doc = sbol2.Document(CRISPR_LOCATION)

        # This returns an object. A TopLevel, not sure what type it is beyond that
        uri = 'http://sbols.org/CRISPR_Example/mKate_gene/1.0.0'
        obj = doc.getExtensionObject(uri)
        self.assertIsNotNone(obj)
        self.assertIsInstance(obj, sbol2.TopLevel)

        # This should raise a not found error because the URI is not
        # contained in the CRISPR test data
        uri = 'http://sbols.org/CRISPR_Example/mKate_genie/1.0.0'
        with self.assertRaises(sbol2.SBOLError) as cm:
            doc.getExtensionObject(uri)
        raised = cm.exception
        self.assertEqual(sbol2.SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                         raised.error_code())

    def test_add_extension_object(self):
        # We need to construct and add an extension object
        # It should then be accessible via getExtensionObject
        # We can also peel back the curtain and make sure the object
        #  is in doc.SBOLObjects and the URI is not in doc.OwnedObjects
        # Three kinds of things to add:
        #  1. An existing type, like ComponentDefinition
        cd = sbol2.ComponentDefinition('cd')
        doc = sbol2.Document()
        doc.addExtensionObject(cd)
        self.assertEqual(1, len(doc.componentDefinitions))
        obj = doc.getExtensionObject(cd.identity)
        self.assertEqual(cd, obj)

        #  2. A new type that is a TopLevel
        tle = TopLevelExtension('tle')
        doc = sbol2.Document()
        doc.addExtensionObject(tle)
        obj = doc.getExtensionObject(tle.identity)
        self.assertEqual(tle, obj)

        #  3. A new type that is not a TopLevel
        # This object won't be found. There is no way for the document to
        # hold on to the object if it isn't a TopLevel
        ntle = NonTopLevelExtension('ntle')
        doc = sbol2.Document()
        doc.addExtensionObject(ntle)
        with self.assertRaises(sbol2.SBOLError) as cm:
            doc.getExtensionObject(ntle.identity)
        raised = cm.exception
        self.assertEqual(sbol2.SBOLErrorCode.SBOL_ERROR_NOT_FOUND,
                         raised.error_code())
        obj = doc.find(ntle.identity)
        self.assertIsNone(obj)

    @unittest.expectedFailure  # See issue #368
    def test_parent_child_extensions(self):
        doc = sbol2.Document()
        tle = TopLevelExtension('tle')
        ntle = NonTopLevelExtension('ntle')
        tle.child = ntle
        doc.add(tle)
        self.assertEqual(len(doc.SBOLObjects), 1)

        # Verify that the parent-child relationship is preserved upon round-trip
        doc.readString(doc.writeString())
        self.assertEqual(len(doc.SBOLObjects), 1)
        self.assertIsNotNone(tle.child)

    def test_register_parent_child_extensions(self):
        # Show that we can load the parent/child relationship when the
        # extension classes are registered. The methods around this one
        # try to do the same thing without registration of the extension
        # classes
        #
        # This is all wrapped in a `try` so that we can clean up the
        # registration in the finally clause
        try:
            sbol2.Config.register_extension_class(TopLevelExtension,
                                                  TopLevelExtension.RDF_TYPE)
            sbol2.Config.register_extension_class(NonTopLevelExtension,
                                                  NonTopLevelExtension.RDF_TYPE)
            doc = sbol2.Document()
            tle = TopLevelExtension('tle')
            ntle = NonTopLevelExtension('ntle')
            tle.child = ntle
            doc.add(tle)
            self.assertEqual(len(doc.SBOLObjects), 1)

            # Verify that the parent-child relationship is preserved upon round-trip
            doc.readString(doc.writeString())
            self.assertEqual(len(doc.SBOLObjects), 1)
            self.assertIsNotNone(tle.child)
        finally:
            # Clean up
            tle_uri = rdflib.URIRef(TopLevelExtension.RDF_TYPE)
            ntle_uri = rdflib.URIRef(NonTopLevelExtension.RDF_TYPE)
            self.assertIn(tle_uri, sbol2.Config.SBOL_DATA_MODEL_REGISTER)
            self.assertIn(ntle_uri, sbol2.Config.SBOL_DATA_MODEL_REGISTER)
            del sbol2.Config.SBOL_DATA_MODEL_REGISTER[tle_uri]
            del sbol2.Config.SBOL_DATA_MODEL_REGISTER[ntle_uri]
            self.assertNotIn(tle_uri, sbol2.Config.SBOL_DATA_MODEL_REGISTER)
            self.assertNotIn(ntle_uri, sbol2.Config.SBOL_DATA_MODEL_REGISTER)

    @unittest.expectedFailure  # See Issue #368
    def test_parent_child_extensions_top_level(self):
        doc = sbol2.Document()
        tle = TopLevelExtension('tle')
        ntle = NonTopLevelExtension('ntle')
        tle.child = ntle
        doc.add(tle)
        self.assertEqual(len(doc.SBOLObjects), 1)

        # Verify that the parent-child relationship is preserved upon round-trip
        doc2 = sbol2.Document()
        doc2.readString(doc.writeString())
        self.assertNotIsInstance(doc2.get(ntle.identity), sbol2.TopLevel)
        self.assertEqual(len(doc.SBOLObjects), 1)
        self.assertIsNotNone(tle.child)

    def test_add_remove_citation(self):
        doc = sbol2.Document()
        sbol1_spec = 'http://www.nature.com/nbt/journal/v32/n6/full/nbt.2891.html'
        sbol2_spec = 'https://doi.org/10.1515/jib-2018-0001'
        sbol3_spec = ('https://sbolstandard.org/wp-content'
                      '/uploads/2020/04/SBOL3.0specification.pdf')
        self.assertEqual([], doc.citations)
        doc.addCitation(sbol1_spec)
        self.assertEqual([sbol1_spec], doc.citations)
        doc.addCitation(sbol2_spec)
        self.assertEqual([sbol1_spec, sbol2_spec], doc.citations)
        doc.addCitation(sbol3_spec)
        expected = [sbol1_spec, sbol2_spec, sbol3_spec]
        self.assertEqual(expected, doc.citations)
        doc.removeCitation(1)
        self.assertEqual([sbol1_spec, sbol3_spec], doc.citations)

    def test_add_remove_keyword(self):
        doc = sbol2.Document()
        keyword1 = 'http://example.com/keyword#key1'
        keyword2 = 'http://example.com/keyword#key2'
        keyword3 = 'http://example.com/keyword#key3'
        self.assertEqual([], doc.keywords)
        doc.addKeyword(keyword1)
        self.assertEqual([keyword1], doc.keywords)
        doc.addKeyword(keyword2)
        self.assertEqual([keyword1, keyword2], doc.keywords)
        doc.addKeyword(keyword3)
        expected = [keyword1, keyword2, keyword3]
        self.assertEqual(expected, doc.keywords)
        doc.removeKeyword(1)
        self.assertEqual([keyword1, keyword3], doc.keywords)

    def test_version(self):
        doc = sbol2.Document()
        self.assertIsNotNone(doc.version)
        old_version = doc.version
        doc.append(CRISPR_LOCATION, overwrite=True)
        self.assertIsNotNone(doc.version)
        self.assertEqual(old_version, doc.version)

    def test_idempotent_read(self):
        doc = sbol2.Document()
        doc.read(CRISPR_LOCATION)
        old_len = len(doc)
        cd_uri = 'http://sbols.org/CRISPR_Example/gRNA_b/1.0.0'
        cd = doc.componentDefinitions[cd_uri]
        self.assertEqual(1, len(cd.roles))

        # Now load the same file again and make sure the size of the
        # document didn't increase, and the number of roles on the
        # ComponentDefinition didn't increase.
        doc.append(CRISPR_LOCATION, overwrite=True)
        self.assertEqual(old_len, len(doc))
        cd = doc.componentDefinitions[cd_uri]
        self.assertEqual(1, len(cd.roles))

        # Now load the file one more time and make sure the size of the
        # document didn't increase, and the number of roles on the
        # ComponentDefinition didn't increase.
        doc.append(CRISPR_LOCATION, overwrite=True)
        self.assertEqual(old_len, len(doc))
        cd = doc.componentDefinitions[cd_uri]
        self.assertEqual(1, len(cd.roles))

    def test_idempotent_read2(self):
        doc = sbol2.Document()
        doc.read(CRISPR_LOCATION)
        old_doc_len = len(doc)
        cd_uri = 'http://sbols.org/CRISPR_Example/gRNA_b/1.0.0'
        cd = doc.componentDefinitions['http://sbols.org/CRISPR_Example/gRNA_b/1.0.0']
        cd.components.create('c')
        old_component_len = len(cd.components)
        with tempfile.TemporaryDirectory() as tmpdirname:
            temp_file = os.path.join(tmpdirname, 'test.xml')
            doc.write(temp_file)
            doc.append(temp_file, overwrite=True)
        cd = doc.componentDefinitions[cd_uri]
        self.assertEqual(old_component_len, len(cd.components))
        self.assertEqual(old_doc_len, len(doc))

    def test_write_validation(self):
        # Test that write performs validation if requested
        # and skips validation if requested.
        doc = sbol2.Document()
        doc.moduleDefinitions.create('md1')
        validate = sbol2.Config.getOption(sbol2.ConfigOptions.VALIDATE)
        sbol2.Config.setOption(sbol2.ConfigOptions.VALIDATE, True)
        verbose = sbol2.Config.getOption(sbol2.ConfigOptions.VERBOSE)
        sbol2.Config.setOption(sbol2.ConfigOptions.VERBOSE, True)
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_file = os.path.join(tmpdirname, 'test.xml')
            with unittest.mock.patch('sys.stdout', new=io.StringIO()) as fake_out:
                # Write to disk
                result = doc.write(test_file)
                self.assertEqual('Valid.', result)
                # Expect timing output
                output = fake_out.getvalue().strip()
                self.assertTrue(output.startswith('Validation request took'))
                self.assertTrue(output.endswith('seconds'))
        sbol2.Config.setOption(sbol2.ConfigOptions.VALIDATE, False)
        with tempfile.TemporaryDirectory() as tmpdirname:
            test_file = os.path.join(tmpdirname, 'test.xml')
            # Write to disk
            result = doc.write(test_file)
            self.assertTrue(result.startswith('Validation disabled.'))
        # Reset validate to its original value
        sbol2.Config.setOption(sbol2.ConfigOptions.VALIDATE, validate)
        sbol2.Config.setOption(sbol2.ConfigOptions.VERBOSE, verbose)

    def test_append_string_no_overwrite(self):
        doc = sbol2.Document(CRISPR_LOCATION)
        with self.assertRaises(sbol2.SBOLError) as cm:
            doc.appendString(doc.writeString(), overwrite=False)
        exc = cm.exception
        self.assertEqual(sbol2.SBOLErrorCode.DUPLICATE_URI_ERROR,
                         exc.error_code())

    def test_append_string(self):
        doc = sbol2.Document()
        cd = doc.componentDefinitions.create('cd1')
        cd.components.create('c1')
        self.assertEqual(1, len(cd.components))
        doc.appendString(doc.writeString(), overwrite=True)
        self.assertEqual(1, len(cd.components))

    def test_append_string_2(self):
        doc = sbol2.Document()
        cd = doc.componentDefinitions.create('cd1')
        cd.components.create('c1')
        cd.roles = ['foo']
        self.assertEqual(1, len(cd.components))
        self.assertEqual('c1', cd.components[0].displayId)
        doc2 = sbol2.Document()
        cd_updated = doc2.componentDefinitions.create('cd1')
        cd_updated.components.create('c2')
        cd_updated.roles = ['bar']
        doc.appendString(doc2.writeString(), overwrite=True)
        self.assertEqual(1, len(cd.components))
        self.assertEqual('c2', cd.components[0].displayId)
        self.assertEqual(cd.roles, ['bar'])

    def test_import_from_format(self):
        genbank_path = os.path.join(MODULE_LOCATION, 'resources', 'brevig-flu.gb')
        doc = sbol2.Document()
        self.assertEqual(0, len(doc))
        doc.importFromFormat(genbank_path)
        self.assertEqual(4, len(doc))
        # Maybe get something out, like a sequence, and make sure
        # it's there and starts with the right stuff.
        self.assertEqual(1, len(doc.sequences))
        display_id = 'AY130766_seq'
        sequence: sbol2.Sequence = doc.getSequence(display_id)
        self.assertTrue(sequence.elements.startswith('atgagtctt'))
        self.assertEqual(982, len(sequence.elements))


if __name__ == '__main__':
    unittest.main()



import re

from .sbolerror import SBOLError, SBOLErrorCode


def is_alphanumeric_or_underscore(c):
    return c.isalnum() or c == '_'


def is_not_alphanumeric_or_underscore(c):
    return not is_alphanumeric_or_underscore(c)


def sbolRule10101(sbol_obj, arg):
    """An SBOL document MUST declare the use of the following XML namespace:
    http://sbols.org/v2#."""
    # TODO
    raise NotImplementedError("Not yet implemented")


def sbolRule10102(sbol_obj, arg):
    """An SBOL document MUST declare the use of the following XML namespace:
    http://www.w3.org/1999/02/22-rdf-syntax-ns#."""
    # TODO
    raise NotImplementedError("Not yet implemented")


def sbol_rule_10202(sbol_obj, arg):
    """The identity property of an Identified object
    MUST be globally unique."""
    if not hasattr(sbol_obj, 'identity'):
        # Not applicable
        raise TypeError('Inappropriate types passed to sbol_rule_10202')
    if not isinstance(arg, str):
        raise TypeError('Inappropriate arg passed to sbol_rule_10202')


# The displayId property of an Identified object is OPTIONAL
# and MAY contain a String that MUST be composed of only alphanumeric or
# underscore characters and MUST NOT begin with a digit.
def sbol_rule_10204(sbol_obj, arg):
    """The displayId property of an Identified object is OPTIONAL and MAY
    contain a String that MUST be composed of only alphanumeric or underscore
    characters and MUST NOT begin with a digit.
    """
    # TODO
    raise NotImplementedError("Not yet implemented")


# The definition property MUST NOT refer to the same ComponentDefinition
# as the one that contains the ComponentInstance. Furthermore,
# ComponentInstance objects MUST NOT form a cyclical chain of references
# via their definition properties and the ComponentDefinition objects that
# contain them. For example, consider the ComponentInstance objects A and B
# and the ComponentDefinition objects X and Y. The reference chain "X
# contains A, A is defined by Y, Y contains B, and B is defined by X"
# is cyclical.


def libsbol_rule_1(sbol_obj, arg):
    """Print a test message"""
    print("Testing internal validation rules")


def libsbol_rule_2(sbol_obj, arg):
    """Validate XSD date-time format"""
    # Implementation note: re.match only matches the beginning
    # part of the string; re.search returns True if any part of the string
    # matches the pattern.
    # Since the intention is to match the entire string,
    # I needed to use '^' and '$'.
    ymd = '[0-9]{4})-([0-9]{2})-([0-9]{2}'
    date_time_1 = re.compile("^(" + ymd + ")([A-Z])?$")
    date_time_2 = re.compile("^(" + ymd + ")-([0-9]{2})T"
                                          "([0-9]{2}):([0-9]{2}):([0-9]{2})"
                                          "([.][0-9]+)?[A-Z]?$")
    date_time_3 = re.compile("^(" + ymd + ")-([0-9]{2})-"
                                          "([0-9]{2})T([0-9]{2}):([0-9]{2}):"
                                          "([0-9]{2})"
                                          "([.][0-9]+)?[A-Z]?"
                                          "([\\+|-]([0-9]{2}):([0-9]{2}))?$")
    datetime_match_1 = re.search(date_time_1, arg) is not None
    datetime_match_2 = re.search(date_time_2, arg) is not None
    datetime_match_3 = re.search(date_time_3, arg) is not None
    if not datetime_match_1 or not datetime_match_2 or not datetime_match_3:
        raise SBOLError(SBOLErrorCode.SBOL_ERROR_NONCOMPLIANT_VERSION,
                        "Invalid datetime format. Datetimes are based on "
                        "XML Schema dateTime datatype. "
                        "For example 2016-03-16T20:12:00Z")


def libsbol_rule_3(sbol_obj, arg):
    raise NotImplementedError("Not yet implemented")


def libsbol_rule_4(sbol_obj, arg):
    raise NotImplementedError("Not yet implemented")


def libsbol_rule_5(sbol_obj, arg):
    raise NotImplementedError("Not yet implemented")


def libsbol_rule_6(sbol_obj, arg):
    raise NotImplementedError("Not yet implemented")


def libsbol_rule_8(sbol_obj, arg):
    raise NotImplementedError("Not yet implemented")


def libsbol_rule_9(sbol_obj, arg):
    raise NotImplementedError("Not yet implemented")


def libsbol_rule_10(sbol_obj, arg):
    raise NotImplementedError("Not yet implemented")


def libsbol_rule_11(sbol_obj, arg):
    # raise NotImplementedError("Not yet implemented")
    pass


def libsbol_rule_12(sbol_obj, arg):
    # raise NotImplementedError("Not yet implemented")
    pass


def libsbol_rule_13(sbol_obj, arg):
    # raise NotImplementedError("Not yet implemented")
    pass


def libsbol_rule_14(sbol_obj, arg):
    # raise NotImplementedError("Not yet implemented")
    pass


def libsbol_rule_15(sbol_obj, arg):
    raise NotImplementedError("Not yet implemented")


def libsbol_rule_16(sbol_obj, arg):
    # raise NotImplementedError("Not yet implemented")
    pass


def libsbol_rule_17(sbol_obj, arg):
    # raise NotImplementedError("Not yet implemented")
    pass


def libsbol_rule_18(sbol_obj, arg):
    # raise NotImplementedError("Not yet implemented")
    pass


def libsbol_rule_22(sbol_obj, arg):
    # raise NotImplementedError("Not yet implemented")
    pass


def libsbol_rule_24(sbol_obj, arg):
    raise NotImplementedError("Not yet implemented")
